# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Elements of a Model
#
# TestSimpleCon                Class for testing single constraint
# TestArrayCon                Class for testing array of constraint
#

import sys
import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.environ import (
    ConcreteModel,
    AbstractModel,
    Var,
    Constraint,
    ConstraintList,
    Param,
    RangeSet,
    Set,
    Expression,
    value,
    simple_constraintlist_rule,
    simple_constraint_rule,
    inequality,
)
from pyomo.common.log import LoggingIntercept
from pyomo.core.expr import (
    SumExpression,
    EqualityExpression,
    InequalityExpression,
    RangedExpression,
)
from pyomo.core.base.constraint import ConstraintData


class TestConstraintCreation(unittest.TestCase):
    def create_model(self, abstract=False):
        if abstract is True:
            model = AbstractModel()
        else:
            model = ConcreteModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        return model

    def test_tuple_construct_equality(self):
        model = self.create_model()

        def rule(model):
            return (0.0, model.x)

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, True)
        self.assertEqual(model.c.lower, 0)
        self.assertIs(model.c.body, model.x)
        self.assertEqual(model.c.upper, 0)

        model = self.create_model()

        def rule(model):
            return (model.x, 0.0)

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, True)
        self.assertEqual(model.c.lower, 0)
        self.assertIs(model.c.body, model.x)
        self.assertEqual(model.c.upper, 0)

    def test_tuple_construct_inf_equality(self):
        model = self.create_model(abstract=True).create_instance()

        model.c = Constraint(expr=(model.x, float('inf')))
        self.assertEqual(model.c.equality, True)
        self.assertEqual(model.c.lower, float('inf'))
        self.assertIs(model.c.body, model.x)
        self.assertEqual(model.c.upper, float('inf'))
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' created with an invalid non-finite lower bound \(inf\).",
        ):
            model.c.lb
        self.assertEqual(model.c.ub, None)

        model.d = Constraint(expr=(float('inf'), model.x))
        self.assertEqual(model.d.equality, True)
        self.assertEqual(model.d.lower, float('inf'))
        self.assertIs(model.d.body, model.x)
        self.assertEqual(model.d.upper, float('inf'))
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'd' created with an invalid non-finite lower bound \(inf\).",
        ):
            model.d.lb
        self.assertEqual(model.d.ub, None)

        model.e = Constraint(expr=(model.x, float('-inf')))
        self.assertEqual(model.e.equality, True)
        self.assertEqual(model.e.lower, float('-inf'))
        self.assertIs(model.e.body, model.x)
        self.assertEqual(model.e.upper, float('-inf'))
        self.assertEqual(model.e.lb, None)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'e' created with an invalid non-finite upper bound \(-inf\).",
        ):
            model.e.ub

        model.f = Constraint(expr=(float('-inf'), model.x))
        self.assertEqual(model.f.equality, True)
        self.assertEqual(model.f.lower, float('-inf'))
        self.assertIs(model.f.body, model.x)
        self.assertEqual(model.f.upper, float('-inf'))
        self.assertEqual(model.f.lb, None)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'f' created with an invalid non-finite upper bound \(-inf\).",
        ):
            model.f.ub

    def test_tuple_construct_1sided_inequality(self):
        model = self.create_model()

        def rule(model):
            return (None, model.y, 1)

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, None)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, 1)

        model = self.create_model()

        def rule(model):
            return (0, model.y, None)

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, 0)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, None)

    def test_tuple_construct_1sided_inf_inequality(self):
        model = self.create_model()

        def rule(model):
            return (float('-inf'), model.y, 1)

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, float('-inf'))
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, 1)
        self.assertEqual(model.c.lb, None)
        self.assertEqual(model.c.ub, 1)

        model = self.create_model()

        def rule(model):
            return (0, model.y, float('inf'))

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, 0)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, float('inf'))
        self.assertEqual(model.c.lb, 0)
        self.assertEqual(model.c.ub, None)

    def test_tuple_construct_unbounded_inequality(self):
        model = self.create_model()

        def rule(model):
            return (None, model.y, None)

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, None)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, None)

        model = self.create_model()

        def rule(model):
            return (float('-inf'), model.y, float('inf'))

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, float('-inf'))
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, float('inf'))
        self.assertEqual(model.c.lb, None)
        self.assertEqual(model.c.ub, None)

    def test_tuple_construct_invalid_1sided_inequality(self):
        model = self.create_model(abstract=True)

        def rule(model):
            return (model.x, model.y, None)

        model.c = Constraint(rule=rule)
        # We now recognize this as a valid inequality
        # self.assertRaises(ValueError, model.create_instance)
        instance = model.create_instance()
        self.assertEqual(instance.c.lower, None)
        self.assertIsInstance(instance.c.body, SumExpression)
        self.assertEqual(instance.c.upper, 0)

        model = self.create_model(abstract=True)

        def rule(model):
            return (None, model.y, model.z)

        model.c = Constraint(rule=rule)
        # We now recognize this as a valid inequality
        # self.assertRaises(ValueError, model.create_instance)
        instance = model.create_instance()
        self.assertEqual(instance.c.lower, None)
        self.assertIsInstance(instance.c.body, SumExpression)
        self.assertEqual(instance.c.upper, 0)

    def test_tuple_construct_2sided_inequality(self):
        model = self.create_model()

        def rule(model):
            return (0, model.y, 1)

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, 0)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, 1)

    def test_tuple_construct_invalid_2sided_inequality(self):
        model = self.create_model(abstract=True)

        def rule(model):
            return (model.x, model.y, 1)

        model.c = Constraint(rule=rule)
        instance = model.create_instance()
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            instance.c.lower
        self.assertIs(instance.c.body, instance.y)
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            instance.c.upper
        instance.x.fix(3)
        self.assertEqual(value(instance.c.lower), 3)

        model = self.create_model(abstract=True)

        def rule(model):
            return (0, model.y, model.z)

        model.c = Constraint(rule=rule)
        instance = model.create_instance()
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable upper bound",
        ):
            instance.c.lower
        self.assertIs(instance.c.body, instance.y)
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable upper bound",
        ):
            instance.c.upper
        instance.z.fix(3)
        self.assertEqual(value(instance.c.upper), 3)

    def test_expr_construct_equality(self):
        model = self.create_model()

        def rule(model):
            return 0.0 == model.x

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, True)
        self.assertEqual(model.c.lower, 0)
        self.assertIs(model.c.body, model.x)
        self.assertEqual(model.c.upper, 0)

        model = self.create_model()

        def rule(model):
            return model.x == 0.0

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, True)
        self.assertEqual(model.c.lower, 0)
        self.assertIs(model.c.body, model.x)
        self.assertEqual(model.c.upper, 0)

    def test_expr_construct_inf_equality(self):
        model = self.create_model(abstract=True).create_instance()

        model.c = Constraint(expr=model.x == float('inf'))
        self.assertEqual(model.c.ub, None)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' created with an invalid non-finite lower bound \(inf\).",
        ):
            model.c.lb

        model.d = Constraint(expr=model.x == float('-inf'))
        self.assertEqual(model.d.lb, None)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'd' created with an invalid non-finite upper bound \(-inf\).",
        ):
            model.d.ub

    def test_expr_construct_1sided_inequality(self):
        model = self.create_model()

        def rule(model):
            return model.y <= 1

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, None)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, 1)

        model = self.create_model()

        def rule(model):
            return 0 <= model.y

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, 0)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, None)

        model = self.create_model()

        def rule(model):
            return model.y >= 1

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, 1)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, None)

        model = self.create_model()

        def rule(model):
            return 0 >= model.y

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, None)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, 0)

    def test_expr_construct_unbounded_inequality(self):
        model = self.create_model()

        def rule(model):
            return model.y <= float('inf')

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertIs(model.c.lower, None)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, float('inf'))
        self.assertIs(model.c.ub, None)
        self.assertIs(model.c.lb, None)

        model = self.create_model()

        def rule(model):
            return float('-inf') <= model.y

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, float('-inf'))
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, None)
        self.assertIs(model.c.ub, None)
        self.assertIs(model.c.lb, None)

        model = self.create_model()

        def rule(model):
            return model.y >= float('-inf')

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, float('-inf'))
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, None)
        self.assertIs(model.c.ub, None)
        self.assertIs(model.c.lb, None)

        model = self.create_model()

        def rule(model):
            return float('inf') >= model.y

        model.c = Constraint(rule=rule)

        self.assertEqual(model.c.equality, False)
        self.assertEqual(model.c.lower, None)
        self.assertIs(model.c.body, model.y)
        self.assertEqual(model.c.upper, float('inf'))
        self.assertIs(model.c.ub, None)
        self.assertIs(model.c.lb, None)

    def test_expr_construct_invalid_unbounded_inequality(self):
        model = self.create_model(abstract=True).create_instance()

        model.c = Constraint(expr=model.y <= float('-inf'))
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' created with an invalid non-finite upper bound \(-inf\).",
        ):
            model.c.ub

        model.d = Constraint(expr=float('inf') <= model.y)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'd' created with an invalid non-finite lower bound \(inf\).",
        ):
            model.d.lb

        model.e = Constraint(expr=model.y >= float('inf'))
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'e' created with an invalid non-finite lower bound \(inf\).",
        ):
            model.e.lb

        model.f = Constraint(expr=float('-inf') >= model.y)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'f' created with an invalid non-finite upper bound \(-inf\).",
        ):
            model.f.ub

    def test_expr_construct_invalid(self):
        m = ConcreteModel()
        c = Constraint(rule=lambda m: None)
        self.assertRaisesRegex(
            ValueError, ".*rule returned None", m.add_component, 'c', c
        )

        m = ConcreteModel()
        c = Constraint([1], rule=lambda m, i: None)
        self.assertRaisesRegex(
            ValueError, ".*rule returned None", m.add_component, 'c', c
        )

        m = ConcreteModel()
        c = Constraint(rule=lambda m: True)
        self.assertRaisesRegex(
            ValueError,
            r".*resolved to a trivial Boolean \(True\).*Constraint\.Feasible",
            m.add_component,
            'c',
            c,
        )

        m = ConcreteModel()
        c = Constraint([1], rule=lambda m, i: True)
        self.assertRaisesRegex(
            ValueError,
            r".*resolved to a trivial Boolean \(True\).*Constraint\.Feasible",
            m.add_component,
            'c',
            c,
        )

        m = ConcreteModel()
        c = Constraint(rule=lambda m: False)
        self.assertRaisesRegex(
            ValueError,
            r".*resolved to a trivial Boolean \(False\).*" r"Constraint\.Infeasible",
            m.add_component,
            'c',
            c,
        )

        m = ConcreteModel()
        c = Constraint([1], rule=lambda m, i: False)
        self.assertRaisesRegex(
            ValueError,
            r".*resolved to a trivial Boolean \(False\).*" r"Constraint\.Infeasible",
            m.add_component,
            'c',
            c,
        )

    def test_nondata_bounds(self):
        model = ConcreteModel()
        model.c = Constraint()
        model.v = Var([1, 2, 3])
        model.e1 = Expression()
        model.e2 = Expression()
        model.e3 = Expression()
        model.c.set_value((model.e1, model.e2, model.e3))
        self.assertIs(model.c.lower, model.e1)
        self.assertIs(model.c.body, model.e2)
        self.assertIs(model.c.upper, model.e3)
        model.e1.expr = 1
        model.e2.expr = 2
        model.e3.expr = 3
        self.assertEqual(value(model.c.lower), 1)
        self.assertEqual(value(model.c.body), 2)
        self.assertEqual(value(model.c.upper), 3)
        model.e1 = model.v[1]
        model.e2 = model.v[2]
        model.e3 = model.v[3]
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            model.c.lower
        self.assertIs(model.c.body.expr, model.v[2])
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            model.c.upper

    # make sure we can use a mutable param that
    # has not been given a value in the upper bound
    # of an inequality constraint
    def test_mutable_novalue_param_lower_bound(self):
        model = ConcreteModel()
        model.x = Var()
        model.p = Param(mutable=True)
        model.p.value = None

        model.c = Constraint(expr=0 <= model.x - model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.p <= model.x)
        self.assertTrue(model.c.lower is model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.p <= model.x + 1)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.p + 1 <= model.x)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.p + 1) ** 2 <= model.x)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.p, model.x, model.p + 1))
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x - model.p >= 0)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x >= model.p)
        self.assertTrue(model.c.lower is model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x + 1 >= model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x >= model.p + 1)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x >= (model.p + 1) ** 2)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.p, model.x, None))
        self.assertTrue(model.c.lower is model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.p, model.x + 1, None))
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.p + 1, model.x, None))
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.p, model.x, 1))
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

    # make sure we can use a mutable param that
    # has not been given a value in the lower bound
    # of an inequality constraint
    def test_mutable_novalue_param_upper_bound(self):
        model = ConcreteModel()
        model.x = Var()
        model.p = Param(mutable=True)
        model.p.value = None

        model.c = Constraint(expr=model.x - model.p <= 0)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x <= model.p)
        self.assertTrue(model.c.upper is model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x + 1 <= model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x <= model.p + 1)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x <= (model.p + 1) ** 2)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.p + 1, model.x, model.p))
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=0 >= model.x - model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.p >= model.x)
        self.assertTrue(model.c.upper is model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.p >= model.x + 1)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=model.p + 1 >= model.x)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.p + 1) ** 2 >= model.x)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(None, model.x, model.p))
        self.assertTrue(model.c.upper is model.p)
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(None, model.x + 1, model.p))
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(None, model.x, model.p + 1))
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

        model.c = Constraint(expr=(1, model.x, model.p))
        self.assertEqual(model.c.equality, False)
        model.del_component(model.c)

    # make sure we can use a mutable param that
    # has not been given a value in the rhs of
    # of an equality constraint
    def test_mutable_novalue_param_equality(self):
        model = ConcreteModel()
        model.x = Var()
        model.p = Param(mutable=True)
        model.p.value = None

        model.c = Constraint(expr=model.x - model.p == 0)
        self.assertEqual(model.c.equality, True)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x == model.p)
        self.assertTrue(model.c.upper is model.p)
        self.assertEqual(model.c.equality, True)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x + 1 == model.p)
        self.assertEqual(model.c.equality, True)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x + 1 == (model.p + 1) ** 2)
        self.assertEqual(model.c.equality, True)
        model.del_component(model.c)

        model.c = Constraint(expr=model.x == model.p + 1)
        self.assertEqual(model.c.equality, True)
        model.del_component(model.c)

        model.c = Constraint(expr=inequality(model.p, model.x, model.p))
        self.assertTrue(model.c.upper is model.p)
        # GH: Not sure if we are supposed to detect equality
        #     in this situation. I would rather us not, for
        #     the sake of making the code less complicated.
        #     Either way, I am not going to test for it here.
        # self.assertEqual(model.c.equality, <blah>)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.x, model.p))
        self.assertTrue(model.c.upper is model.p)
        self.assertEqual(model.c.equality, True)
        model.del_component(model.c)

        model.c = Constraint(expr=(model.p, model.x))
        self.assertTrue(model.c.upper is model.p)
        self.assertEqual(model.c.equality, True)
        model.del_component(model.c)

    def test_inequality(self):
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(expr=inequality(lower=-1, body=m.x))
        self.assertEqual(m.c.lower.value, -1)
        self.assertIs(m.c.body, m.x)
        self.assertIs(m.c.upper, None)

        del m.c
        m.c = Constraint(expr=inequality(body=m.x, upper=1))
        self.assertIs(m.c.lower, None)
        self.assertIs(m.c.body, m.x)
        self.assertEqual(m.c.upper.value, 1)

        del m.c
        m.c = Constraint(expr=inequality(lower=-1, body=m.x, upper=1))
        self.assertEqual(m.c.lower.value, -1)
        self.assertIs(m.c.body, m.x)
        self.assertEqual(m.c.upper.value, 1)


class TestSimpleCon(unittest.TestCase):
    def test_set_expr_explicit_multivariate(self):
        """Test expr= option (multivariate expression)"""
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.x = Var(model.A, initialize=2)
        ans = 0
        for i in model.A:
            ans = ans + model.x[i]
        ans = ans >= 0
        ans = ans <= 1
        model.c = Constraint(expr=ans)

        self.assertEqual(model.c(), 8)
        self.assertEqual(model.c.body(), 8)
        self.assertEqual(value(model.c.body), 8)

    def test_set_expr_explicit_univariate(self):
        """Test expr= option (univariate expression)"""
        model = ConcreteModel()
        model.x = Var(initialize=2)
        ans = model.x >= 0
        ans = ans <= 1
        model.c = Constraint(expr=ans)

        self.assertEqual(model.c(), 2)
        self.assertEqual(model.c.body(), 2)
        self.assertEqual(value(model.c.body), 2)

    def test_set_expr_undefined_univariate(self):
        """Test expr= option (univariate expression)"""
        model = ConcreteModel()
        model.x = Var()
        ans = model.x >= 0
        ans = ans <= 1
        model.c = Constraint(expr=ans)

        with self.assertRaisesRegex(
            ValueError, "No value for uninitialized ScalarVar object x"
        ):
            value(model.c)
        model.x = 2
        self.assertEqual(model.c(), 2)
        self.assertEqual(value(model.c.body), 2)

    def test_set_expr_inline(self):
        """Test expr= option (inline expression)"""
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.x = Var(model.A, initialize=2)
        model.c = Constraint(expr=(0, sum(model.x[i] for i in model.A), 1))

        self.assertEqual(model.c(), 8)
        self.assertEqual(value(model.c.body), 8)

    def test_rule1(self):
        """Test rule option"""
        model = ConcreteModel()
        model.B = RangeSet(1, 4)

        def f(model):
            ans = 0
            for i in model.B:
                ans = ans + model.x[i]
            ans = ans >= 0
            ans = ans <= 1
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(rule=f)

        self.assertEqual(model.c(), 8)
        self.assertEqual(value(model.c.body), 8)

    def test_rule2(self):
        """Test rule option"""
        model = ConcreteModel()
        model.B = RangeSet(1, 4)

        def f(model):
            ans = 0
            for i in model.B:
                ans = ans + model.x[i]
            return (0, ans, 1)

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(rule=f)

        self.assertEqual(model.c(), 8)
        self.assertEqual(value(model.c.body), 8)

    def test_rule3(self):
        """Test rule option"""
        model = ConcreteModel()
        model.B = RangeSet(1, 4)

        def f(model):
            ans = 0
            for i in model.B:
                ans = ans + model.x[i]
            return (0, ans, None)

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(rule=f)

        self.assertEqual(model.c(), 8)
        self.assertEqual(value(model.c.body), 8)

    def test_rule4(self):
        """Test rule option"""
        model = ConcreteModel()
        model.B = RangeSet(1, 4)

        def f(model):
            ans = 0
            for i in model.B:
                ans = ans + model.x[i]
            return (None, ans, 1)

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(rule=f)

        self.assertEqual(model.c(), 8)
        self.assertEqual(value(model.c.body), 8)

    def test_rule5(self):
        """Test rule option"""
        model = ConcreteModel()
        model.B = RangeSet(1, 4)

        def f(model):
            ans = 0
            for i in model.B:
                ans = ans + model.x[i]
            return (ans, 1)

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(rule=f)

        self.assertEqual(model.c(), 8)
        self.assertEqual(value(model.c.body), 8)

    def test_dim(self):
        """Test dim method"""
        model = ConcreteModel()
        model.c = Constraint()

        self.assertEqual(model.c.dim(), 0)

    def test_keys_empty(self):
        """Test keys method"""
        model = ConcreteModel()
        model.c = Constraint()

        self.assertEqual(list(model.c.keys()), [])

    def test_len_empty(self):
        """Test len method"""
        model = ConcreteModel()
        model.c = Constraint()

        self.assertEqual(len(model.c), 0)

    def test_None_key(self):
        """Test keys method"""
        model = ConcreteModel()
        model.x = Var()
        model.c = Constraint(expr=model.x == 1)
        self.assertEqual(list(model.c.keys()), [None])
        self.assertEqual(id(model.c), id(model.c[None]))

    def test_len(self):
        """Test len method"""
        model = AbstractModel()
        model.x = Var()
        model.c = Constraint(rule=lambda m: m.x == 1)
        self.assertEqual(len(model.c), 0)
        inst = model.create_instance()
        self.assertEqual(len(inst.c), 1)

    def test_setitem(self):
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint()
        self.assertEqual(len(m.c), 0)

        m.c = m.x**2 <= 4
        self.assertEqual(len(m.c), 1)
        self.assertEqual(list(m.c.keys()), [None])
        self.assertEqual(m.c.upper, 4)

        m.c = Constraint.Skip
        self.assertEqual(len(m.c), 0)


class TestArrayCon(unittest.TestCase):
    def create_model(self):
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2, 3, 4])
        return model

    def test_rule_option1(self):
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i):
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(model.A, rule=f)

        self.assertEqual(model.c[1](), 8)
        self.assertEqual(model.c[2](), 16)
        self.assertEqual(len(model.c), 4)

    def test_rule_option2(self):
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i):
            if i % 2 == 0:
                return Constraint.Skip
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(model.A, rule=f)

        self.assertEqual(model.c[1](), 8)
        self.assertEqual(len(model.c), 2)

    def test_rule_option3(self):
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i):
            if i % 2 == 0:
                return Constraint.Skip
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(model.A, rule=f)

        self.assertEqual(model.c[1](), 8)
        self.assertEqual(len(model.c), 2)

    def test_rule_option2a(self):
        model = self.create_model()
        model.B = RangeSet(1, 4)

        @simple_constraint_rule
        def f(model, i):
            if i % 2 == 0:
                return None
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(model.A, rule=f)

        self.assertEqual(model.c[1](), 8)
        self.assertEqual(len(model.c), 2)

    def test_rule_option3a(self):
        model = self.create_model()
        model.B = RangeSet(1, 4)

        @simple_constraint_rule
        def f(model, i):
            if i % 2 == 0:
                return None
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(model.A, rule=f)

        self.assertEqual(model.c[1](), 8)
        self.assertEqual(len(model.c), 2)

    def test_dim(self):
        model = self.create_model()
        model.c = Constraint(model.A)

        self.assertEqual(model.c.dim(), 1)

    def test_keys(self):
        model = self.create_model()
        model.c = Constraint(model.A)

        self.assertEqual(len(list(model.c.keys())), 0)

    def test_len(self):
        model = self.create_model()
        model.c = Constraint(model.A)
        self.assertEqual(len(model.c), 0)

        model = self.create_model()
        model.B = RangeSet(1, 4)
        """Test rule option"""

        def f(model):
            ans = 0
            for i in model.B:
                ans = ans + model.x[i]
            ans = ans == 2
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(rule=f)

        self.assertEqual(len(model.c), 1)

    def test_setitem(self):
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(range(5))
        self.assertEqual(len(m.c), 0)

        m.c[2] = m.x**2 <= 4
        self.assertEqual(len(m.c), 1)
        self.assertEqual(list(m.c.keys()), [2])
        self.assertIsInstance(m.c[2], ConstraintData)
        self.assertEqual(m.c[2].upper, 4)

        m.c[3] = Constraint.Skip
        self.assertEqual(len(m.c), 1)
        self.assertRaisesRegex(KeyError, "3", m.c.__getitem__, 3)

        self.assertRaisesRegex(
            ValueError, r"'c\[3\]': rule returned None", m.c.__setitem__, 3, None
        )
        self.assertEqual(len(m.c), 1)

        m.c[2] = Constraint.Skip
        self.assertEqual(len(m.c), 0)


class TestConList(unittest.TestCase):
    def create_model(self):
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2, 3, 4])
        return model

    #
    # Tests that adding Constraint.Skip increments
    # the internal counter but does not create an object
    #
    def test_conlist_skip(self):
        model = ConcreteModel()
        model.x = Var()
        model.c = ConstraintList()
        self.assertTrue(1 not in model.c)
        self.assertEqual(len(model.c), 0)
        model.c.add(Constraint.Skip)
        self.assertTrue(1 not in model.c)
        self.assertEqual(len(model.c), 0)
        model.c.add(model.x >= 1)
        self.assertTrue(1 not in model.c)
        self.assertTrue(2 in model.c)
        self.assertEqual(len(model.c), 1)

    def test_rule_option1(self):
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i):
            if i > 4:
                return ConstraintList.End
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = ConstraintList(rule=f)

        self.assertEqual(model.c[1](), 8)
        self.assertEqual(model.c[2](), 16)
        self.assertEqual(len(model.c), 4)

    def test_rule_option2(self):
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i):
            if i > 2:
                return ConstraintList.End
            i = 2 * i - 1
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = ConstraintList(rule=f)

        self.assertEqual(model.c[1](), 8)
        self.assertEqual(len(model.c), 2)

    def test_rule_option1a(self):
        model = self.create_model()
        model.B = RangeSet(1, 4)

        @simple_constraintlist_rule
        def f(model, i):
            if i > 4:
                return None
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = ConstraintList(rule=f)

        self.assertEqual(model.c[1](), 8)
        self.assertEqual(model.c[2](), 16)
        self.assertEqual(len(model.c), 4)

    def test_rule_option2a(self):
        model = self.create_model()
        model.B = RangeSet(1, 4)

        @simple_constraintlist_rule
        def f(model, i):
            if i > 2:
                return None
            i = 2 * i - 1
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = ConstraintList(rule=f)

        self.assertEqual(model.c[1](), 8)
        self.assertEqual(len(model.c), 2)

    def test_rule_option3(self):
        model = self.create_model()
        model.y = Var(initialize=2)

        def f(model):
            yield model.y <= 0
            yield 2 * model.y <= 0
            yield 2 * model.y <= 0
            yield ConstraintList.End

        model.c = ConstraintList(rule=f)
        self.assertEqual(len(model.c), 3)
        self.assertEqual(model.c[1](), 2)
        model.d = ConstraintList(rule=f(model))
        self.assertEqual(len(model.d), 3)
        self.assertEqual(model.d[1](), 2)

    def test_rule_option4(self):
        model = self.create_model()
        model.y = Var(initialize=2)
        model.c = ConstraintList(rule=((i + 1) * model.y >= 0 for i in range(3)))
        self.assertEqual(len(model.c), 3)
        self.assertEqual(model.c[1](), 2)

    def test_dim(self):
        model = self.create_model()
        model.c = ConstraintList()

        self.assertEqual(model.c.dim(), 1)

    def test_keys(self):
        model = self.create_model()
        model.c = ConstraintList()

        self.assertEqual(len(list(model.c.keys())), 0)

    def test_len(self):
        model = self.create_model()
        model.c = ConstraintList()

        self.assertEqual(len(model.c), 0)

    def test_0based_add(self):
        m = ConcreteModel()
        m.x = Var()
        m.c = ConstraintList(starting_index=0)
        m.c.add(m.x <= 0)
        self.assertEqual(list(m.c.keys()), [0])
        m.c.add(m.x >= 0)
        self.assertEqual(list(m.c.keys()), [0, 1])


class Test2DArrayCon(unittest.TestCase):
    def create_model(self):
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2])
        return model

    def test_rule_option(self):
        """Test rule option"""
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i, j):
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            ans = ans <= 0
            ans = ans >= 0
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(model.A, model.A, rule=f)

        self.assertEqual(model.c[1, 1](), 8)
        self.assertEqual(model.c[2, 1](), 16)

    def test_dim(self):
        """Test dim method"""
        model = self.create_model()
        model.c = Constraint(model.A, model.A)

        self.assertEqual(model.c.dim(), 2)

    def test_keys(self):
        """Test keys method"""
        model = self.create_model()
        model.c = Constraint(model.A, model.A)

        self.assertEqual(len(list(model.c.keys())), 0)

    def test_len(self):
        """Test len method"""
        model = self.create_model()
        model.c = Constraint(model.A, model.A)
        self.assertEqual(len(model.c), 0)

        model = self.create_model()
        model.B = RangeSet(1, 4)
        """Test rule option"""

        def f(model):
            ans = 0
            for i in model.B:
                ans = ans + model.x[i]
            ans = ans == 2
            return ans

        model.x = Var(model.B, initialize=2)
        model.c = Constraint(rule=f)

        self.assertEqual(len(model.c), 1)


class MiscConTests(unittest.TestCase):
    def test_infeasible(self):
        m = ConcreteModel()
        m.c = Constraint(expr=Constraint.Infeasible)
        self.assertIn(None, m.c._data)
        self.assertEqual(m.c.lb, None)
        self.assertEqual(m.c.body, 1)
        self.assertEqual(m.c.ub, 0)

        m.c = (0, 1, 2)
        self.assertIn(None, m.c._data)
        self.assertEqual(m.c.lb, 0)
        self.assertEqual(m.c.body, 1)
        self.assertEqual(m.c.ub, 2)

        m.c = Constraint.Infeasible
        self.assertIn(None, m.c._data)
        self.assertEqual(m.c.lb, None)
        self.assertEqual(m.c.body, 1)
        self.assertEqual(m.c.ub, 0)

    def test_feasible(self):
        m = ConcreteModel()
        m.c = Constraint(expr=Constraint.Feasible)
        self.assertIn(None, m.c._data)
        self.assertEqual(m.c.lb, None)
        self.assertEqual(m.c.body, 0)
        self.assertEqual(m.c.ub, 0)

        m.c = (0, 1, 2)
        self.assertIn(None, m.c._data)
        self.assertEqual(m.c.lb, 0)
        self.assertEqual(m.c.body, 1)
        self.assertEqual(m.c.ub, 2)

        m.c = Constraint.Feasible
        self.assertIn(None, m.c._data)
        self.assertEqual(m.c.lb, None)
        self.assertEqual(m.c.body, 0)
        self.assertEqual(m.c.ub, 0)

    def test_slack_methods(self):
        model = ConcreteModel()
        model.x = Var(initialize=2.0)
        L = -1.0
        U = 5.0
        model.cL = Constraint(expr=model.x**2 >= L)
        self.assertEqual(model.cL.lslack(), 5.0)
        self.assertEqual(model.cL.uslack(), float('inf'))
        self.assertEqual(model.cL.slack(), 5.0)
        model.cU = Constraint(expr=model.x**2 <= U)
        self.assertEqual(model.cU.lslack(), float('inf'))
        self.assertEqual(model.cU.uslack(), 1.0)
        self.assertEqual(model.cU.slack(), 1.0)
        model.cR = Constraint(expr=(L, model.x**2, U))
        self.assertEqual(model.cR.lslack(), 5.0)
        self.assertEqual(model.cR.uslack(), 1.0)
        self.assertEqual(model.cR.slack(), 1.0)

    def test_constructor(self):
        a = Constraint(name="b")
        self.assertEqual(a.local_name, "b")
        try:
            a = Constraint(foo="bar")
            self.fail("Can't specify an unexpected constructor option")
        except ValueError:
            pass

    def test_contains(self):
        model = ConcreteModel()
        model.a = Set(initialize=[1, 2, 3])
        model.b = Constraint(model.a)

        self.assertEqual(2 in model.b, False)
        tmp = []
        for i in model.b:
            tmp.append(i)
        self.assertEqual(len(tmp), 0)

    def test_empty_singleton(self):
        a = Constraint()
        a.construct()
        #
        # Even though we construct a ScalarConstraint,
        # if it is not initialized that means it is "empty"
        # and we should encounter errors when trying to access the
        # ConstraintData interface methods until we assign
        # something to the constraint.
        #
        self.assertEqual(a._constructed, True)
        self.assertEqual(len(a), 0)
        try:
            a()
            self.fail("Component is empty")
        except ValueError:
            pass
        try:
            a.body
            self.fail("Component is empty")
        except ValueError:
            pass
        try:
            a.lower
            self.fail("Component is empty")
        except ValueError:
            pass
        try:
            a.upper
            self.fail("Component is empty")
        except ValueError:
            pass
        try:
            a.equality
            self.fail("Component is empty")
        except ValueError:
            pass
        try:
            a.strict_lower
            self.fail("Component is empty")
        except ValueError:
            pass
        try:
            a.strict_upper
            self.fail("Component is empty")
        except ValueError:
            pass
        x = Var(initialize=1.0)
        x.construct()
        a.set_value((0, x, 2))
        self.assertEqual(len(a), 1)
        self.assertEqual(a(), 1)
        self.assertEqual(a.body(), 1)
        self.assertEqual(a.lower(), 0)
        self.assertEqual(a.upper(), 2)
        self.assertEqual(a.equality, False)
        self.assertEqual(a.strict_lower, False)
        self.assertEqual(a.strict_upper, False)

    def test_unconstructed_singleton(self):
        a = Constraint()
        self.assertEqual(a._constructed, False)
        self.assertEqual(len(a), 0)
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access .* on AbstractScalarConstraint"
            ".*before it has been constructed",
        ):
            a()
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access .* on AbstractScalarConstraint"
            ".*before it has been constructed",
        ):
            a.body
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access .* on AbstractScalarConstraint"
            ".*before it has been constructed",
        ):
            a.lower
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access .* on AbstractScalarConstraint"
            ".*before it has been constructed",
        ):
            a.upper
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access .* on AbstractScalarConstraint"
            ".*before it has been constructed",
        ):
            a.equality
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access .* on AbstractScalarConstraint"
            ".*before it has been constructed",
        ):
            a.strict_lower
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access .* on AbstractScalarConstraint"
            ".*before it has been constructed",
        ):
            a.strict_upper

        x = Var(initialize=1.0)
        x.construct()
        a.construct()
        a.set_value((0, x, 2))
        self.assertEqual(len(a), 1)
        self.assertEqual(a(), 1)
        self.assertEqual(a.body(), 1)
        self.assertEqual(a.lower(), 0)
        self.assertEqual(a.upper(), 2)
        self.assertEqual(a.equality, False)
        self.assertEqual(a.strict_lower, False)
        self.assertEqual(a.strict_upper, False)

    def test_deprecated_rule_attribute(self):
        def rule(m):
            return m.x <= 0

        def new_rule(m):
            return m.x >= 0

        m = ConcreteModel()
        m.x = Var()
        m.con = Constraint(rule=rule)

        self.assertIs(m.con.rule._fcn, rule)
        with LoggingIntercept() as LOG:
            m.con.rule = new_rule
        self.assertIn(
            "DEPRECATED: The 'Constraint.rule' attribute will be made read-only",
            LOG.getvalue(),
        )
        self.assertIs(m.con.rule, new_rule)

    def test_rule(self):
        def rule1(model):
            return Constraint.Skip

        model = ConcreteModel()
        try:
            model.o = Constraint(rule=rule1)
        except Exception:
            e = sys.exc_info()[1]
            self.fail("Failure to create empty constraint: %s" % str(e))

        #
        def rule1(model):
            return (0.0, model.x, 2.0)

        model = ConcreteModel()
        model.x = Var(initialize=1.1)
        model.o = Constraint(rule=rule1)

        self.assertEqual(model.o(), 1.1)

        #
        def rule1(model, i):
            return Constraint.Skip

        model = ConcreteModel()
        model.a = Set(initialize=[1, 2, 3])
        try:
            model.o = Constraint(model.a, rule=rule1)
        except Exception:
            self.fail("Error generating empty constraint")

        #
        def rule1(model):
            return (0.0, 1.1, 2.0, None)

        model = ConcreteModel()
        try:
            model.o = Constraint(rule=rule1)
            self.fail("Can only return tuples of length 2 or 3")
        except ValueError:
            pass

    def test_rule_kwargs(self):
        m = ConcreteModel()
        m.x = Var()

        @m.Constraint(rhs=5)
        def c(m, *, rhs):
            return m.x <= rhs

        self.assertExpressionsEqual(m.c.expr, m.x <= 5)

    def test_tuple_constraint_create(self):
        def rule1(model):
            return (0.0, model.x)

        model = ConcreteModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.c = Constraint(rule=rule1)
        self.assertEqual(model.c.lower, 0)
        self.assertIs(model.c.body, model.x)
        self.assertEqual(model.c.upper, 0)

        #
        def rule1(model):
            return (model.y, model.x, model.z)

        model = AbstractModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.c = Constraint(rule=rule1)
        instance = model.create_instance()
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            instance.c.lower
        self.assertIs(instance.c.body, instance.x)
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            instance.c.upper

        #
        def rule1(model):
            return (0, model.x, model.z)

        model = AbstractModel()
        model.x = Var()
        model.z = Var()
        model.c = Constraint(rule=rule1)
        instance = model.create_instance()
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable upper bound",
        ):
            instance.c.lower
        self.assertIs(instance.c.body, instance.x)
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' is a Ranged Inequality with a variable upper bound",
        ):
            instance.c.upper

    def test_expression_constructor_coverage(self):
        def rule1(model):
            expr = model.x
            expr = expr == 0.0
            expr = expr >= 1.0
            return expr

        model = AbstractModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.o = Constraint(rule=rule1)
        self.assertRaises(TypeError, model.create_instance)

        #
        def rule1(model):
            expr = model.U >= model.x
            expr = expr >= model.L
            return expr

        model = ConcreteModel()
        model.x = Var()
        model.L = Param(initialize=0)
        model.U = Param(initialize=1)
        model.o = Constraint(rule=rule1)

        #
        def rule1(model):
            expr = model.x <= model.z
            expr = expr >= model.y
            return expr

        model = AbstractModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.o = Constraint(rule=rule1)

        # self.assertRaises(ValueError, model.create_instance)
        #
        def rule1(model):
            expr = model.x >= model.z
            expr = model.y >= expr
            return expr

        model = AbstractModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.o = Constraint(rule=rule1)

        # self.assertRaises(ValueError, model.create_instance)
        #
        def rule1(model):
            expr = model.y <= model.x
            expr = model.y >= expr
            return expr

        model = AbstractModel()
        model.x = Var()
        model.y = Var()
        model.o = Constraint(rule=rule1)

        # self.assertRaises(ValueError, model.create_instance)
        #
        def rule1(model):
            expr = model.x >= model.L
            return expr

        model = ConcreteModel()
        model.x = Var()
        model.L = Param(initialize=0)
        model.o = Constraint(rule=rule1)

        #
        def rule1(model):
            expr = model.U >= model.x
            return expr

        model = ConcreteModel()
        model.x = Var()
        model.U = Param(initialize=0)
        model.o = Constraint(rule=rule1)

        #
        def rule1(model):
            expr = model.x
            expr = expr == 0.0
            expr = expr <= 1.0
            return expr

        model = AbstractModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.o = Constraint(rule=rule1)
        self.assertRaises(TypeError, model.create_instance)

        #
        def rule1(model):
            expr = model.U <= model.x
            expr = expr <= model.L
            return expr

        model = ConcreteModel()
        model.x = Var()
        model.L = Param(initialize=0)
        model.U = Param(initialize=1)
        model.o = Constraint(rule=rule1)

        #
        def rule1(model):
            expr = model.x >= model.z
            expr = expr <= model.y
            return expr

        model = AbstractModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.o = Constraint(rule=rule1)

        # self.assertRaises(ValueError, model.create_instance)
        #
        def rule1(model):
            expr = model.x <= model.z
            expr = model.y <= expr
            return expr

        model = AbstractModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.o = Constraint(rule=rule1)

        # self.assertRaises(ValueError, model.create_instance)
        #
        def rule1(model):
            expr = model.x <= model.L
            return expr

        model = ConcreteModel()
        model.x = Var()
        model.L = Param(initialize=0)
        model.o = Constraint(rule=rule1)

        #
        def rule1(model):
            expr = model.y >= model.x
            expr = model.y <= expr
            return expr

        model = AbstractModel()
        model.x = Var()
        model.y = Var()
        model.o = Constraint(rule=rule1)

        # self.assertRaises(ValueError, model.create_instance)
        #
        def rule1(model):
            expr = model.U <= model.x
            return expr

        model = ConcreteModel()
        model.x = Var()
        model.U = Param(initialize=0)
        model.o = Constraint(rule=rule1)

        #
        def rule1(model):
            return model.x + model.x

        model = ConcreteModel()
        model.x = Var()
        try:
            model.o = Constraint(rule=rule1)
            self.fail("Cannot return an unbounded expression")
        except ValueError:
            pass
        #

    def test_abstract_index(self):
        model = AbstractModel()
        model.A = Set()
        model.B = Set()
        model.C = model.A | model.B
        model.x = Constraint(model.C)

    def test_ranged_inequality_expr(self):
        model = ConcreteModel()
        model.v = Var()
        model.l = Param(initialize=1, mutable=True)
        model.u = Param(initialize=3, mutable=True)
        model.con = Constraint(expr=inequality(model.l, model.v, model.u))
        self.assertIs(model.con.expr.args[0], model.l)
        self.assertIs(model.con.expr.args[1], model.v)
        self.assertIs(model.con.expr.args[2], model.u)

    def test_potentially_variable_bounds(self):
        m = ConcreteModel()
        m.x = Var()
        m.l = Expression()
        m.u = Expression()
        m.c = Constraint(expr=inequality(m.l, m.x, m.u))
        self.assertIs(m.c.lower, m.l)
        self.assertIs(m.c.upper, m.u)
        with self.assertRaisesRegex(
            ValueError, 'No value for uninitialized ScalarExpression object l'
        ):
            m.c.lb
        with self.assertRaisesRegex(
            ValueError, 'No value for uninitialized ScalarExpression object u'
        ):
            m.c.ub

        m.l = 5
        m.u = 10
        self.assertIs(m.c.lower, m.l)
        self.assertIs(m.c.upper, m.u)
        self.assertEqual(m.c.lb, 5)
        self.assertEqual(m.c.ub, 10)

        m.l.expr = m.x
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            m.c.lower
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            self.assertIs(m.c.upper, m.u)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            m.c.lb
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' is a Ranged Inequality with a variable lower bound",
        ):
            self.assertEqual(m.c.ub, 10)

        m.l = 15
        m.u.expr = m.x
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' is a Ranged Inequality with a variable upper bound",
        ):
            self.assertIs(m.c.lower, m.l)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' is a Ranged Inequality with a variable upper bound",
        ):
            m.c.upper
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' is a Ranged Inequality with a variable upper bound",
        ):
            self.assertEqual(m.c.lb, 15)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' is a Ranged Inequality with a variable upper bound",
        ):
            m.c.ub

        m.l = -float('inf')
        m.u = float('inf')
        self.assertIs(m.c.lower, m.l)
        self.assertIs(m.c.upper, m.u)
        self.assertIsNone(m.c.lb)
        self.assertIsNone(m.c.ub)

        m.l = float('inf')
        m.u = -float('inf')
        self.assertIs(m.c.lower, m.l)
        self.assertIs(m.c.upper, m.u)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' created with an invalid "
            r"non-finite lower bound \(inf\)",
        ):
            m.c.lb
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' created with an invalid "
            r"non-finite upper bound \(-inf\)",
        ):
            m.c.ub

        m.l = float('nan')
        m.u = -float('nan')
        self.assertIs(m.c.lower, m.l)
        self.assertIs(m.c.upper, m.u)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' created with an invalid "
            r"non-finite lower bound \(nan\)",
        ):
            m.c.lb
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' created with an invalid "
            r"non-finite upper bound \(nan\)",
        ):
            m.c.ub

    def test_tuple_expression(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.p = Param(mutable=True, initialize=0)
        m.c = Constraint()

        m.c = (m.x, m.y)
        self.assertTrue(m.c.equality)
        self.assertIs(type(m.c.expr), EqualityExpression)

        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'c' does not have a proper value. "
            "Equality Constraints expressed as 2-tuples cannot "
            "contain None",
        ):
            m.c = (m.x, None)

        # You can create it with an infinite value, but then one of the
        # bounds will fail:
        m.c = (m.x, float('inf'))
        self.assertIsNone(m.c.ub)
        with self.assertRaisesRegex(
            ValueError,
            r"Constraint 'c' created with an invalid "
            r"non-finite lower bound \(inf\)",
        ):
            m.c.lb


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.sympy_tools import (
    sympy_available,
    sympyify_expression,
    sympy2pyomo_expression,
)

# A "public" attribute indicating that differentiate() can be called
# ... this provides a bit of future-proofing for alternative approaches
# to symbolic differentiation.
differentiate_available = sympy_available


def differentiate(expr, wrt=None, wrt_list=None):
    """Return derivative of expression.

    This function returns an expression or list of expression objects
    corresponding to the derivative of the passed expression 'expr' with
    respect to a variable 'wrt' or list of variables 'wrt_list'

    Args:
        expr (Expression): Pyomo expression
        wrt (Var): Pyomo variable
        wrt_list (list): list of Pyomo variables

    Returns:
        Expression or list of Expression objects

    """
    if not sympy_available:
        raise RuntimeError(
            "The sympy module is not available.\n\t"
            "Cannot perform automatic symbolic differentiation."
        )
    if not ((wrt is None) ^ (wrt_list is None)):
        raise ValueError(
            "differentiate(): Must specify exactly one of wrt and wrt_list"
        )
    import sympy

    #
    # Convert the Pyomo expression to a sympy expression
    #
    objectMap, sympy_expr = sympyify_expression(expr)
    #
    # The partial_derivs dict holds intermediate sympy expressions that
    # we can re-use.  We will prepopulate it with None for all vars that
    # appear in the expression (so that we can detect wrt combinations
    # that are, by definition, 0)
    #
    partial_derivs = {x: None for x in objectMap.sympyVars()}
    #
    # Setup the WRT list
    #
    if wrt is not None:
        wrt_list = [wrt]
    else:
        # Copy the list because we will normalize things in place below
        wrt_list = list(wrt_list)
    #
    # Convert WRT vars into sympy vars
    #
    ans = [None] * len(wrt_list)
    for i, target in enumerate(wrt_list):
        if target.__class__ is not tuple:
            target = (target,)
        wrt_list[i] = tuple(objectMap.getSympySymbol(x) for x in target)
        for x in wrt_list[i]:
            if x not in partial_derivs:
                ans[i] = 0.0
                break
    #
    # We assume that users will not request duplicate derivatives.  We
    # will only cache up to the next-to last partial, and if a user
    # requests the exact same derivative twice, then we will just
    # re-calculate it.
    #
    last_partial_idx = max(len(x) for x in wrt_list) - 1
    #
    # Calculate all the derivatives
    #
    for i, target in enumerate(wrt_list):
        if ans[i] is not None:
            continue
        part = sympy_expr
        for j, wrt_var in enumerate(target):
            if j == last_partial_idx:
                part = sympy.diff(part, wrt_var)
            else:
                partial_target = target[: j + 1]
                if partial_target in partial_derivs:
                    part = partial_derivs[partial_target]
                else:
                    part = sympy.diff(part, wrt_var)
                    partial_derivs[partial_target] = part
        ans[i] = sympy2pyomo_expression(part, objectMap)
    #
    # Return the answer
    #
    return ans if wrt is None else ans[0]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Provides functions to compute and enforce disjunctive variable bounds.

These are tighter variable bounds that are valid within the scope of a certain
disjunct. That is, these are bounds on variables that hold true when an
associated disjunct is active (indicator_var value is 1). The bounds are
enforced using a ConstraintList on every Disjunct, enforcing the relevant
variable bounds. This may lead to duplication of constraints, so the constraints
to variable bounds preprocessing transformation is recommended for NLP problems
processed with this transformation.

"""

from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.common.errors import InfeasibleConstraintException
from pyomo.contrib.fbbt.fbbt import fbbt, BoundsManager
from pyomo.core.base.block import Block, TraversalStrategy
from pyomo.core.expr import identify_variables
from pyomo.core import Constraint, Objective, TransformationFactory, minimize, value
from pyomo.opt import SolverFactory
from pyomo.gdp.disjunct import Disjunct
from pyomo.core.plugins.transform.hierarchy import Transformation
from pyomo.opt import TerminationCondition as tc

linear_degrees = {0, 1}
inf = float('inf')


def disjunctive_obbt(model, solver):
    """Provides Optimality-based bounds tightening to a model using a solver."""
    model._disjuncts_to_process = list(
        model.component_data_objects(
            ctype=Disjunct,
            active=True,
            descend_into=(Block, Disjunct),
            descent_order=TraversalStrategy.BreadthFirstSearch,
        )
    )
    if model.ctype == Disjunct:
        model._disjuncts_to_process.insert(0, model)

    linear_var_set = ComponentSet()
    for constr in model.component_data_objects(
        Constraint, active=True, descend_into=(Block, Disjunct)
    ):
        if constr.body.polynomial_degree() in linear_degrees:
            linear_var_set.update(identify_variables(constr.body, include_fixed=False))
    model._disj_bnds_linear_vars = list(linear_var_set)

    for disj_idx, disjunct in enumerate(model._disjuncts_to_process):
        var_bnds = obbt_disjunct(model, disj_idx, solver)
        if var_bnds is not None:
            # Add bounds to the disjunct
            if not hasattr(disjunct, '_disj_var_bounds'):
                # No bounds had been computed before. Attach the bounds dictionary.
                disjunct._disj_var_bounds = var_bnds
            else:
                # Update the bounds dictionary.
                for var, new_bnds in var_bnds.items():
                    old_lb, old_ub = disjunct._disj_var_bounds.get(var, (-inf, inf))
                    new_lb, new_ub = new_bnds
                    disjunct._disj_var_bounds[var] = (
                        max(old_lb, new_lb),
                        min(old_ub, new_ub),
                    )
        else:
            disjunct.deactivate()  # prune disjunct


def obbt_disjunct(orig_model, idx, solver):
    model = orig_model.clone()

    # Fix the disjunct to be active
    disjunct = model._disjuncts_to_process[idx]
    disjunct.indicator_var.fix(1)

    for obj in model.component_data_objects(Objective, active=True):
        obj.deactivate()

    # Deactivate nonlinear constraints
    for constr in model.component_data_objects(
        Constraint, active=True, descend_into=(Block, Disjunct)
    ):
        if constr.body.polynomial_degree() not in linear_degrees:
            constr.deactivate()

    # Only look at the variables participating in active constraints within the scope
    relevant_var_set = ComponentSet()
    for constr in disjunct.component_data_objects(Constraint, active=True):
        relevant_var_set.update(identify_variables(constr.body, include_fixed=False))

    TransformationFactory('gdp.bigm').apply_to(model)

    model._var_bounding_obj = Objective(expr=1, sense=minimize)

    for var in relevant_var_set:
        model._var_bounding_obj.set_value(expr=var)
        var_lb = solve_bounding_problem(model, solver)
        if var_lb is None:
            return None  # bounding problem infeasible
        model._var_bounding_obj.set_value(expr=-var)
        var_ub = solve_bounding_problem(model, solver)
        if var_ub is None:
            return None  # bounding problem infeasible
        else:
            var_ub = -var_ub  # sign correction

        var.setlb(var_lb)
        var.setub(var_ub)

    # Maps original variable --> (new computed LB, new computed UB)
    var_bnds = ComponentMap(
        (
            (
                orig_var,
                (
                    clone_var.lb if clone_var.has_lb() else -inf,
                    clone_var.ub if clone_var.has_ub() else inf,
                ),
            )
            for orig_var, clone_var in zip(
                orig_model._disj_bnds_linear_vars, model._disj_bnds_linear_vars
            )
            if clone_var in relevant_var_set
        )
    )
    return var_bnds


def solve_bounding_problem(model, solver):
    results = SolverFactory(solver).solve(model)
    if results.solver.termination_condition is tc.optimal:
        return value(model._var_bounding_obj.expr)
    elif results.solver.termination_condition is tc.infeasible:
        return None
    elif results.solver.termination_condition is tc.unbounded:
        return -inf
    else:
        raise NotImplementedError(
            "Unhandled termination condition: %s" % results.solver.termination_condition
        )


def disjunctive_fbbt(model):
    """Applies FBBT to a model"""
    fbbt_disjunct(model, ComponentMap())


def fbbt_disjunct(disj, parent_bounds):
    orig_bnds = ComponentMap(parent_bounds)
    try:
        for var, var_bnds in disj._disj_var_bounds.items():
            scope_lb, scope_ub = var_bnds
            scope_lb = -inf if scope_lb is None else scope_lb
            scope_ub = inf if scope_ub is None else scope_ub
            parent_lb, parent_ub = parent_bounds.get(var, (-inf, inf))
            orig_bnds[var] = (max(scope_lb, parent_lb), min(scope_ub, parent_ub))
    except AttributeError:
        # disj._disj_var_bounds does not exist yet
        pass
    bnds_manager = BoundsManager(disj)
    bnds_manager.load_bounds(orig_bnds)
    try:
        new_bnds = fbbt(disj)
    except InfeasibleConstraintException as e:
        if disj.ctype == Disjunct:
            disj.deactivate()  # simply prune the disjunct
        new_bnds = parent_bounds
    bnds_manager.pop_bounds()
    disj._disj_var_bounds = new_bnds
    # Handle nested disjuncts
    for disj in disj.component_data_objects(Disjunct, active=True):
        fbbt_disjunct(disj, new_bnds)


@TransformationFactory.register(
    'contrib.compute_disj_var_bounds',
    doc="Compute disjunctive bounds in a given model.",
)
class ComputeDisjunctiveVarBounds(Transformation):
    """Compute disjunctive bounds in a given model.

    Tries to compute the disjunctive bounds for all variables found in
    constraints that are in disjuncts under the given model.

    Two strategies are available to compute the disjunctive bounds:
     - Feasibility-based bounds tightening using the contrib.fbbt package. (Default)
     - Optimality-based bounds tightening by solving the linear relaxation of the model.

    This transformation introduces ComponentMap objects named _disj_var_bounds to
    each Disjunct and the top-level model object. These map var --> (var.disj_lb, var.disj_ub)
    for each disjunctive scope.

    Args:
        model (Component): The model under which to look for disjuncts.
        solver (string): The solver to use for OBBT, or None for FBBT.

    """

    def _apply_to(self, model, solver=None):
        """Apply the transformation.

        Args:
            model: Pyomo model object on which to compute disjuctive bounds.

        """
        if solver is not None:
            disjunctive_obbt(model, solver)
        else:
            disjunctive_fbbt(model)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests the induced linearity module."""

import pyomo.common.unittest as unittest
from pyomo.contrib.preprocessing.plugins.induced_linearity import (
    _bilinear_expressions,
    detect_effectively_discrete_vars,
    determine_valid_values,
)
from pyomo.common.collections import ComponentSet, Bunch
from pyomo.environ import (
    Binary,
    ConcreteModel,
    Constraint,
    ConstraintList,
    Integers,
    RangeSet,
    SolverFactory,
    TransformationFactory,
    Var,
    exp,
)
from pyomo.gdp import Disjunct, Disjunction
from pyomo.repn import generate_standard_repn

glpk_available = SolverFactory('glpk').available()


class TestInducedLinearity(unittest.TestCase):
    """Tests induced linearity."""

    def test_detect_bilinear_vars(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.z = Var()
        m.c = Constraint(
            expr=(m.x - 3) * (m.y + 2)
            - (m.z + 4) * m.y
            + (m.x + 2) ** 2
            + exp(m.y**2) * m.x
            <= m.z
        )
        m.c2 = Constraint(expr=m.x * m.y == 3)
        bilinear_map = _bilinear_expressions(m)
        self.assertEqual(len(bilinear_map), 3)
        self.assertEqual(len(bilinear_map[m.x]), 2)
        self.assertEqual(len(bilinear_map[m.y]), 2)
        self.assertEqual(len(bilinear_map[m.z]), 1)
        self.assertEqual(bilinear_map[m.x][m.x], ComponentSet([m.c]))
        self.assertEqual(bilinear_map[m.x][m.y], ComponentSet([m.c, m.c2]))
        self.assertEqual(bilinear_map[m.y][m.x], ComponentSet([m.c, m.c2]))
        self.assertEqual(bilinear_map[m.y][m.z], ComponentSet([m.c]))
        self.assertEqual(bilinear_map[m.z][m.y], ComponentSet([m.c]))

    def test_detect_effectively_discrete_vars(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var(domain=Binary)
        m.z = Var(domain=Integers)
        m.constr = Constraint(expr=m.x == m.y + m.z)
        m.ignore_inequality = Constraint(expr=m.x <= m.y + m.z)
        m.ignore_nonlinear = Constraint(expr=m.x**2 == m.y + m.z)
        m.a = Var()
        m.b = Var(domain=Binary)
        m.c = Var(domain=Integers)
        m.disj = Disjunct()
        m.disj.constr = Constraint(expr=m.a == m.b + m.c)
        effectively_discrete = detect_effectively_discrete_vars(m, 1e-6)
        self.assertEqual(len(effectively_discrete), 1)
        self.assertEqual(effectively_discrete[m.x], [m.constr])
        effectively_discrete = detect_effectively_discrete_vars(m.disj, 1e-6)
        self.assertEqual(len(effectively_discrete), 1)
        self.assertEqual(effectively_discrete[m.a], [m.disj.constr])

    @unittest.skipIf(not glpk_available, 'GLPK not available')
    def test_determine_valid_values(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var(RangeSet(4), domain=Binary)
        m.z = Var(domain=Integers, bounds=(-1, 2))
        m.constr = Constraint(
            expr=m.x == m.y[1] + 2 * m.y[2] + m.y[3] + 2 * m.y[4] + m.z
        )
        m.logical = ConstraintList()
        m.logical.add(expr=m.y[1] + m.y[2] == 1)
        m.logical.add(expr=m.y[3] + m.y[4] == 1)
        m.logical.add(expr=m.y[2] + m.y[4] <= 1)
        var_to_values_map = determine_valid_values(
            m,
            detect_effectively_discrete_vars(m, 1e-6),
            Bunch(equality_tolerance=1e-6, pruning_solver='glpk'),
        )
        valid_values = set([1, 2, 3, 4, 5])
        self.assertEqual(set(var_to_values_map[m.x]), valid_values)

    @unittest.skipIf(not glpk_available, 'GLPK not available')
    def test_induced_linearity_case2(self):
        m = ConcreteModel()
        m.x = Var([0], bounds=(-3, 8))
        m.y = Var(RangeSet(4), domain=Binary)
        m.z = Var(domain=Integers, bounds=(-1, 2))
        m.constr = Constraint(
            expr=m.x[0] == m.y[1] + 2 * m.y[2] + m.y[3] + 2 * m.y[4] + m.z
        )
        m.logical = ConstraintList()
        m.logical.add(expr=m.y[1] + m.y[2] == 1)
        m.logical.add(expr=m.y[3] + m.y[4] == 1)
        m.logical.add(expr=m.y[2] + m.y[4] <= 1)
        m.b = Var(bounds=(-2, 7))
        m.c = Var()
        m.bilinear = Constraint(
            expr=(m.x[0] - 3) * (m.b + 2) - (m.c + 4) * m.b + exp(m.b**2) * m.x[0]
            <= m.c
        )
        TransformationFactory('contrib.induced_linearity').apply_to(m)
        xfrmed_blk = m._induced_linearity_info.x0_b_bilinear
        self.assertSetEqual(set(xfrmed_blk.valid_values), set([1, 2, 3, 4, 5]))
        select_one_repn = generate_standard_repn(xfrmed_blk.select_one_value.body)
        self.assertEqual(
            ComponentSet(select_one_repn.linear_vars),
            ComponentSet(xfrmed_blk.x_active[i] for i in xfrmed_blk.valid_values),
        )

    @unittest.skipIf(not glpk_available, 'GLPK not available')
    def test_bilinear_in_disjuncts(self):
        m = ConcreteModel()
        m.x = Var([0], bounds=(-3, 8))
        m.y = Var(RangeSet(4), domain=Binary)
        m.z = Var(domain=Integers, bounds=(-1, 2))
        m.constr = Constraint(
            expr=m.x[0] == m.y[1] + 2 * m.y[2] + m.y[3] + 2 * m.y[4] + m.z
        )
        m.logical = ConstraintList()
        m.logical.add(expr=m.y[1] + m.y[2] == 1)
        m.logical.add(expr=m.y[3] + m.y[4] == 1)
        m.logical.add(expr=m.y[2] + m.y[4] <= 1)
        m.v = Var([1, 2])
        m.v[1].setlb(-2)
        m.v[1].setub(7)
        m.v[2].setlb(-4)
        m.v[2].setub(5)
        m.bilinear = Constraint(
            expr=(m.x[0] - 3) * (m.v[1] + 2)
            - (m.v[2] + 4) * m.v[1]
            + exp(m.v[1] ** 2) * m.x[0]
            <= m.v[2]
        )
        m.disjctn = Disjunction(expr=[[m.x[0] * m.v[1] <= 4], [m.x[0] * m.v[2] >= 6]])
        TransformationFactory('contrib.induced_linearity').apply_to(m)
        self.assertEqual(
            m.disjctn.disjuncts[0].constraint[1].body.polynomial_degree(), 1
        )
        self.assertEqual(
            m.disjctn.disjuncts[1].constraint[1].body.polynomial_degree(), 1
        )

    @unittest.skipIf(not glpk_available, 'GLPK not available')
    def test_induced_linear_in_disjunct(self):
        m = ConcreteModel()
        m.x = Var([0], bounds=(-3, 8))
        m.y = Var(RangeSet(2), domain=Binary)
        m.logical = ConstraintList()
        m.logical.add(expr=m.y[1] + m.y[2] == 1)
        m.v = Var([1])
        m.v[1].setlb(-2)
        m.v[1].setub(7)
        m.bilinear_outside = Constraint(expr=m.x[0] * m.v[1] >= 2)
        m.disjctn = Disjunction(
            expr=[
                [m.x[0] * m.v[1] == 3, 2 * m.x[0] == m.y[1] + m.y[2]],
                [m.x[0] * m.v[1] == 4],
            ]
        )
        TransformationFactory('contrib.induced_linearity').apply_to(m)
        self.assertEqual(
            m.disjctn.disjuncts[0].constraint[1].body.polynomial_degree(), 1
        )
        self.assertEqual(m.bilinear_outside.body.polynomial_degree(), 2)
        self.assertEqual(
            m.disjctn.disjuncts[1].constraint[1].body.polynomial_degree(), 2
        )


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
This module contains transformations for representing a
single-variate piecewise linear function using a
mixed-integer problem formulation (see [VAN10]_).

"""

import logging
import bisect

# TODO: Figure out of the 'log' and 'dlog' representations
# really do require (2^n)+1 points or if there is a way to
# handle the between sizes.

from pyomo.core.expr.numvalue import value as _value
from pyomo.core.kernel.set_types import IntegerSet
from pyomo.core.kernel.block import block
from pyomo.core.kernel.expression import expression, expression_tuple
from pyomo.core.kernel.variable import (
    IVariable,
    variable_list,
    variable_tuple,
    variable_dict,
    variable,
)
from pyomo.core.kernel.constraint import (
    constraint_list,
    constraint_tuple,
    linear_constraint,
)
from pyomo.core.kernel.sos import sos2
from pyomo.core.kernel.piecewise_library.util import (
    characterize_function,
    is_nondecreasing,
    is_positive_power_of_two,
    log2floor,
    generate_gray_code,
    PiecewiseValidationError,
)

logger = logging.getLogger('pyomo.core')

registered_transforms = {}


# wrapper that allows a list containing parameters to be
# used with the bisect module
class _shadow_list:
    __slots__ = ("_x",)

    def __init__(self, x):
        self._x = x

    def __len__(self):
        return self._x.__len__()

    def __getitem__(self, i):
        return _value(self._x.__getitem__(i))


def piecewise(
    breakpoints,
    values,
    input=None,
    output=None,
    bound='eq',
    repn='sos2',
    validate=True,
    simplify=True,
    equal_slopes_tolerance=1e-6,
    require_bounded_input_variable=True,
    require_variable_domain_coverage=True,
):
    """
    Models a single-variate piecewise linear function.

    This function takes a list breakpoints and function
    values describing a piecewise linear function and
    transforms this input data into a block of variables and
    constraints that enforce a piecewise linear relationship
    between an input variable and an output variable. In the
    general case, this transformation requires the use of
    discrete decision variables.

    Args:
        breakpoints (list): The list of breakpoints of the
            piecewise linear function. This can be a list of
            numbers or a list of objects that store mutable
            data (e.g., mutable parameters). If mutable data
            is used validation might need to be disabled by
            setting the :attr:`validate` keyword to
            :const:`False`. The list of breakpoints must be
            in non-decreasing order.
        values (list): The values of the piecewise linear
            function corresponding to the breakpoints.
        input: The variable constrained to be the input of
            the piecewise linear function.
        output: The variable constrained to be the output of
            the piecewise linear function.
        bound (str): The type of bound to impose on the
            output expression. Can be one of:

              - 'lb': y <= f(x)
              - 'eq': y  = f(x)
              - 'ub': y >= f(x)
        repn (str): The type of piecewise representation to
            use. Choices are shown below (+ means step
            functions are supported)

                - 'sos2': standard representation using sos2 constraints (+)
                -  'dcc': disaggregated convex combination (+)
                - 'dlog': logarithmic disaggregated convex combination (+)
                -   'cc': convex combination (+)
                -  'log': logarithmic branching convex combination (+)
                -   'mc': multiple choice
                -  'inc': incremental method (+)
        validate (bool): Indicates whether or not to perform
            validation of the input data. The default is
            :const:`True`. Validation can be performed
            manually after the piecewise object is created
            by calling the :meth:`validate`
            method. Validation should be performed any time
            the inputs are changed (e.g., when using mutable
            parameters in the breakpoints list or when the
            input variable changes).
        simplify (bool): Indicates whether or not to attempt
            to simplify the piecewise representation to
            avoid using discrete variables. This can be done
            when the feasible region for the output
            variable, with respect to the piecewise function
            and the bound type, is a convex set. Default is
            :const:`True`. Validation is required to perform
            simplification, so this keyword is ignored when
            the :attr:`validate` keyword is :attr:`False`.
        equal_slopes_tolerance (float): Tolerance used check
            if consecutive slopes are nearly equal. If any
            are found, validation will fail. Default is
            1e-6. This keyword is ignored when the
            :attr:`validate` keyword is :attr:`False`.
        require_bounded_input_variable (bool): Indicates if
            the input variable is required to have finite
            upper and lower bounds. Default is
            :const:`True`. Setting this keyword to
            :const:`False` can be used to allow general
            expressions to be used as the input in place of
            a variable. This keyword is ignored when the
            :attr:`validate` keyword is :attr:`False`.
        require_variable_domain_coverage (bool): Indicates
            if the function domain (defined by the endpoints
            of the breakpoints list) needs to cover the
            entire domain of the input variable. Default is
            :const:`True`. Ignored for any bounds of
            variables that are not finite, or when the input
            is not assigned a variable. This keyword is
            ignored when the :attr:`validate` keyword is
            :attr:`False`.

    Returns:
        TransformedPiecewiseLinearFunction: a block that
            stores any new variables, constraints, and other
            modeling objects used by the piecewise representation
    """
    transform = None
    try:
        transform = registered_transforms[repn]
    except KeyError:
        raise ValueError(
            "Keyword assignment repn='%s' is not valid. "
            "Must be one of: %s" % (repn, str(sorted(registered_transforms.keys())))
        )
    assert transform is not None

    if not validate:
        # can not simplify if we do not validate
        simplify = False

    func = PiecewiseLinearFunction(breakpoints, values, validate=False)

    if simplify and (transform is not piecewise_convex):
        ftype = func.validate(equal_slopes_tolerance=equal_slopes_tolerance)

        if (bound == 'eq') and (ftype == characterize_function.affine):
            transform = piecewise_convex
        elif (bound == 'lb') and (
            ftype in (characterize_function.affine, characterize_function.convex)
        ):
            transform = piecewise_convex
        elif (bound == 'ub') and (
            ftype in (characterize_function.affine, characterize_function.concave)
        ):
            transform = piecewise_convex

    return transform(
        func,
        input=input,
        output=output,
        bound=bound,
        validate=validate,
        equal_slopes_tolerance=equal_slopes_tolerance,
        require_bounded_input_variable=require_bounded_input_variable,
        require_variable_domain_coverage=require_variable_domain_coverage,
    )


class PiecewiseLinearFunction:
    """A piecewise linear function

    Piecewise linear functions are defined by a list of
    breakpoints and a list function values corresponding to
    each breakpoint. The function value between breakpoints
    is implied through linear interpolation.

    Args:
        breakpoints (list): The list of function
            breakpoints.
        values (list): The list of function values (one for
            each breakpoint).
        validate (bool): Indicates whether or not to perform
            validation of the input data. The default is
            :const:`True`. Validation can be performed
            manually after the piecewise object is created
            by calling the :meth:`validate`
            method. Validation should be performed any time
            the inputs are changed (e.g., when using mutable
            parameters in the breakpoints list).
        **kwds: Additional keywords are passed to the
            :meth:`validate` method when the :attr:`validate`
            keyword is :const:`True`; otherwise, they are
            ignored.
    """

    __slots__ = ("_breakpoints", "_values")

    def __init__(self, breakpoints, values, validate=True, **kwds):
        self._breakpoints = breakpoints
        self._values = values
        if type(self._breakpoints) is not tuple:
            self._breakpoints = tuple(self._breakpoints)
        if type(self._values) is not tuple:
            self._values = tuple(self._values)
        if len(self._breakpoints) != len(self._values):
            raise ValueError(
                "The number of breakpoints (%s) differs from "
                "the number of function values (%s)"
                % (len(self._breakpoints), len(self._values))
            )
        if validate:
            self.validate(**kwds)

    def __getstate__(self):
        """Required for older versions of the pickle
        protocol since this class uses __slots__"""
        return {key: getattr(self, key) for key in self.__slots__}

    def __setstate__(self, state):
        """Required for older versions of the pickle
        protocol since this class uses __slots__"""
        for key in state:
            setattr(self, key, state[key])

    def validate(self, equal_slopes_tolerance=1e-6):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints and values
        lists (e.g., that the list of breakpoints is
        nondecreasing).

        Args:
            equal_slopes_tolerance (float): Tolerance used
                check if consecutive slopes are nearly
                equal. If any are found, validation will
                fail. Default is 1e-6.

        Returns:
            int:
                a function characterization code (see
                :func:`util.characterize_function`)

        Raises:
            PiecewiseValidationError: if validation fails
        """

        breakpoints = [_value(x) for x in self._breakpoints]
        values = [_value(x) for x in self._values]
        if not is_nondecreasing(breakpoints):
            raise PiecewiseValidationError(
                "The list of breakpoints is not nondecreasing: %s" % (str(breakpoints))
            )

        ftype, slopes = characterize_function(breakpoints, values)
        for i in range(1, len(slopes)):
            if (
                (slopes[i - 1] is not None)
                and (slopes[i] is not None)
                and (abs(slopes[i - 1] - slopes[i]) <= equal_slopes_tolerance)
            ):
                raise PiecewiseValidationError(
                    "Piecewise function validation detected slopes "
                    "of consecutive line segments to be within %s "
                    "of one another. This may cause numerical issues. "
                    "To avoid this error, set the 'equal_slopes_tolerance' "
                    "keyword to a smaller value or disable validation."
                    % (equal_slopes_tolerance)
                )

        return ftype

    @property
    def breakpoints(self):
        """The set of breakpoints used to defined this function"""
        return self._breakpoints

    @property
    def values(self):
        """The set of values used to defined this function"""
        return self._values

    def __call__(self, x):
        """Evaluates the piecewise linear function at the
        given point using interpolation. Note that step functions are
        assumed lower-semicontinuous."""
        i = bisect.bisect_left(_shadow_list(self.breakpoints), x)
        if i == 0:
            xP = _value(self.breakpoints[i])
            if xP == x:
                return float(_value(self.values[i]))
        elif i != len(self.breakpoints):
            xL = _value(self.breakpoints[i - 1])
            xU = _value(self.breakpoints[i])
            assert xL <= xU
            if (xL <= x) and (x <= xU):
                yL = _value(self.values[i - 1])
                yU = _value(self.values[i])
                return yL + (float(yU - yL) / (xU - xL)) * (x - xL)
        raise ValueError(
            "The point %s is outside of the "
            "function domain: [%s,%s]."
            % (x, _value(self.breakpoints[0]), _value(self.breakpoints[-1]))
        )


class TransformedPiecewiseLinearFunction(block):
    """Base class for transformed piecewise linear functions

    A transformed piecewise linear functions is a block of
    variables and constraints that enforce a piecewise
    linear relationship between an input variable and an
    output variable.

    Args:
        f (:class:`PiecewiseLinearFunction`): The piecewise
            linear function to transform.
        input: The variable constrained to be the input of
            the piecewise linear function.
        output: The variable constrained to be the output of
            the piecewise linear function.
        bound (str): The type of bound to impose on the
            output expression. Can be one of:

              - 'lb': y <= f(x)
              - 'eq': y  = f(x)
              - 'ub': y >= f(x)
        validate (bool): Indicates whether or not to perform
            validation of the input data. The default is
            :const:`True`. Validation can be performed
            manually after the piecewise object is created
            by calling the :meth:`validate`
            method. Validation should be performed any time
            the inputs are changed (e.g., when using mutable
            parameters in the breakpoints list or when the
            input variable changes).
        **kwds: Additional keywords are passed to the
            :meth:`validate` method when the :attr:`validate`
            keyword is :const:`True`; otherwise, they are
            ignored.
    """

    def __init__(self, f, input=None, output=None, bound='eq', validate=True, **kwds):
        super(TransformedPiecewiseLinearFunction, self).__init__()
        assert isinstance(f, PiecewiseLinearFunction)
        if bound not in ('lb', 'ub', 'eq'):
            raise ValueError(
                "Invalid bound type %r. Must be one of: ['lb','ub','eq']" % (bound)
            )
        self._bound = bound
        self._f = f
        self._inout = expression_tuple([expression(input), expression(output)])
        if validate:
            self.validate(**kwds)

    @property
    def input(self):
        """The expression that stores the input to the
        piecewise function. The returned object can be
        updated by assigning to its :attr:`expr`
        attribute."""
        return self._inout[0]

    @property
    def output(self):
        """The expression that stores the output of the
        piecewise function. The returned object can be
        updated by assigning to its :attr:`expr`
        attribute."""
        return self._inout[1]

    @property
    def bound(self):
        """The bound type assigned to the piecewise
        relationship ('lb','ub','eq')."""
        return self._bound

    def validate(
        self,
        equal_slopes_tolerance=1e-6,
        require_bounded_input_variable=True,
        require_variable_domain_coverage=True,
    ):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints, values, and
        input variable (e.g., that the list of breakpoints
        is nondecreasing).

        Args:
            equal_slopes_tolerance (float): Tolerance used
                check if consecutive slopes are nearly
                equal. If any are found, validation will
                fail. Default is 1e-6.
            require_bounded_input_variable (bool): Indicates
                if the input variable is required to have
                finite upper and lower bounds. Default is
                :const:`True`. Setting this keyword to
                :const:`False` can be used to allow general
                expressions to be used as the input in place
                of a variable.
            require_variable_domain_coverage (bool):
                Indicates if the function domain (defined by
                the endpoints of the breakpoints list) needs
                to cover the entire domain of the input
                variable. Default is :const:`True`. Ignored
                for any bounds of variables that are not
                finite, or when the input is not assigned a
                variable.

        Returns:
            int:
                a function characterization code (see
                :func:`util.characterize_function`)

        Raises:
            PiecewiseValidationError: if validation fails
        """
        ftype = self._f.validate(equal_slopes_tolerance=equal_slopes_tolerance)
        assert ftype in (1, 2, 3, 4, 5)

        input_var = self.input.expr
        if not isinstance(input_var, IVariable):
            input_var = None

        if require_bounded_input_variable and (
            (input_var is None) or (not input_var.has_lb()) or (not input_var.has_ub())
        ):
            raise PiecewiseValidationError(
                "Piecewise function input is not a "
                "variable with finite upper and lower "
                "bounds: %s. To avoid this error, set the "
                "'require_bounded_input_variable' keyword "
                "to False or disable validation." % (str(input_var))
            )

        if require_variable_domain_coverage and (input_var is not None):
            domain_lb = _value(self.breakpoints[0])
            domain_ub = _value(self.breakpoints[-1])
            if input_var.has_lb() and _value(input_var.lb) < domain_lb:
                raise PiecewiseValidationError(
                    "Piecewise function domain does not include "
                    "the lower bound of the input variable: "
                    "%s.ub = %s > %s. To avoid this error, set "
                    "the 'require_variable_domain_coverage' "
                    "keyword to False or disable validation."
                    % (input_var.name, _value(input_var.lb), domain_lb)
                )
            if input_var.has_ub() and _value(input_var.ub) > domain_ub:
                raise PiecewiseValidationError(
                    "Piecewise function domain does not include "
                    "the upper bound of the input variable: "
                    "%s.ub = %s > %s. To avoid this error, set "
                    "the 'require_variable_domain_coverage' "
                    "keyword to False or disable validation."
                    % (input_var.name, _value(input_var.ub), domain_ub)
                )

        return ftype

    @property
    def breakpoints(self):
        """The set of breakpoints used to defined this function"""
        return self._f.breakpoints

    @property
    def values(self):
        """The set of values used to defined this function"""
        return self._f.values

    def __call__(self, x):
        """Evaluates the piecewise linear function at the
        given point using interpolation"""
        return self._f(x)


class piecewise_convex(TransformedPiecewiseLinearFunction):
    """Simple convex piecewise representation

    Expresses a piecewise linear function with a convex
    feasible region for the output variable using a simple
    collection of linear constraints.
    """

    def __init__(self, *args, **kwds):
        super(piecewise_convex, self).__init__(*args, **kwds)

        breakpoints = self.breakpoints
        values = self.values
        self.c = constraint_list()
        for i in range(len(breakpoints) - 1):
            X0 = breakpoints[i]
            F_AT_X0 = values[i]
            dF_AT_X0 = (values[i + 1] - F_AT_X0) / (breakpoints[i + 1] - X0)
            const = F_AT_X0 - dF_AT_X0 * X0
            con = linear_constraint((self.output, self.input), (-1, dF_AT_X0))
            if self.bound == 'ub':
                con.lb = -const
            elif self.bound == 'lb':
                con.ub = -const
            else:
                assert self.bound == 'eq'
                con.rhs = -const
            self.c.append(con)

        # In order to enforce the same behavior as actual
        # piecewise constraints, we need to constrain the
        # input expression to be between first and last
        # breakpoint. This might be duplicating the
        # variable, but its not always the case, and there's
        # no guarantee that the input "variable" is not a
        # more general linear expression.
        self.c.append(
            linear_constraint(
                terms=[(self.input, 1)], lb=self.breakpoints[0], ub=self.breakpoints[-1]
            )
        )

    def validate(self, **kwds):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints, values, and
        input variable (e.g., that the list of breakpoints
        is nondecreasing).

        See base class documentation for keyword
        descriptions.
        """
        ftype = super(piecewise_convex, self).validate(**kwds)
        if (self.bound == 'eq') and (ftype != characterize_function.affine):
            raise PiecewiseValidationError(
                "The bound type is 'eq' but the function "
                "was not characterized as affine (only two "
                "breakpoints). The 'convex' piecewise "
                "representation does not support this function."
            )
        elif (self.bound == 'lb') and (
            ftype not in (characterize_function.affine, characterize_function.convex)
        ):
            raise PiecewiseValidationError(
                "The bound type is 'lb' but the function "
                "was not characterized as convex or affine. "
                "The 'convex' piecewise representation does "
                "not support this function."
            )
        elif (self.bound == 'ub') and (
            ftype not in (characterize_function.affine, characterize_function.concave)
        ):
            raise PiecewiseValidationError(
                "The bound type is 'ub' but the function "
                "was not characterized as concave or affine. "
                "The 'convex' piecewise representation does "
                "not support this function."
            )
        return ftype


registered_transforms['convex'] = piecewise_convex


class piecewise_sos2(TransformedPiecewiseLinearFunction):
    """Discrete SOS2 piecewise representation

    Expresses a piecewise linear function using
    the SOS2 formulation.
    """

    def __init__(self, *args, **kwds):
        super(piecewise_sos2, self).__init__(*args, **kwds)

        # create vars
        y_tuple = tuple(variable(lb=0) for i in range(len(self.breakpoints)))
        y = self.v = variable_tuple(y_tuple)

        # create piecewise constraints
        self.c = constraint_list()

        self.c.append(
            linear_constraint(
                variables=y_tuple + (self.input,),
                coefficients=self.breakpoints + (-1,),
                rhs=0,
            )
        )

        self.c.append(
            linear_constraint(
                variables=y_tuple + (self.output,), coefficients=self.values + (-1,)
            )
        )
        if self.bound == 'ub':
            self.c[-1].lb = 0
        elif self.bound == 'lb':
            self.c[-1].ub = 0
        else:
            assert self.bound == 'eq'
            self.c[-1].rhs = 0

        self.c.append(
            linear_constraint(variables=y_tuple, coefficients=(1,) * len(y), rhs=1)
        )

        self.s = sos2(y)

    def validate(self, **kwds):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints, values, and
        input variable (e.g., that the list of breakpoints
        is nondecreasing).

        See base class documentation for keyword
        descriptions.
        """
        return super(piecewise_sos2, self).validate(**kwds)


registered_transforms['sos2'] = piecewise_sos2


class piecewise_dcc(TransformedPiecewiseLinearFunction):
    """Discrete DCC piecewise representation

    Expresses a piecewise linear function using
    the DCC formulation.
    """

    def __init__(self, *args, **kwds):
        super(piecewise_dcc, self).__init__(*args, **kwds)

        # create index sets
        polytopes = range(len(self.breakpoints) - 1)
        vertices = range(len(self.breakpoints))

        def polytope_verts(p):
            return range(p, p + 2)

        # create vars
        self.v = variable_dict()
        lmbda = self.v['lambda'] = variable_dict(
            ((p, v), variable(lb=0)) for p in polytopes for v in vertices
        )
        y = self.v['y'] = variable_tuple(
            variable(domain_type=IntegerSet, lb=0, ub=1) for p in polytopes
        )

        # create piecewise constraints
        self.c = constraint_list()

        self.c.append(
            linear_constraint(
                variables=tuple(
                    lmbda[p, v] for p in polytopes for v in polytope_verts(p)
                )
                + (self.input,),
                coefficients=tuple(
                    self.breakpoints[v] for p in polytopes for v in polytope_verts(p)
                )
                + (-1,),
                rhs=0,
            )
        )

        self.c.append(
            linear_constraint(
                variables=tuple(
                    lmbda[p, v] for p in polytopes for v in polytope_verts(p)
                )
                + (self.output,),
                coefficients=tuple(
                    self.values[v] for p in polytopes for v in polytope_verts(p)
                )
                + (-1,),
            )
        )
        if self.bound == 'ub':
            self.c[-1].lb = 0
        elif self.bound == 'lb':
            self.c[-1].ub = 0
        else:
            assert self.bound == 'eq'
            self.c[-1].rhs = 0

        clist = []
        for p in polytopes:
            variables = tuple(lmbda[p, v] for v in polytope_verts(p))
            clist.append(
                linear_constraint(
                    variables=variables + (y[p],),
                    coefficients=(1,) * len(variables) + (-1,),
                    rhs=0,
                )
            )
        self.c.append(constraint_tuple(clist))

        self.c.append(
            linear_constraint(variables=tuple(y), coefficients=(1,) * len(y), rhs=1)
        )

    def validate(self, **kwds):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints, values, and
        input variable (e.g., that the list of breakpoints
        is nondecreasing).

        See base class documentation for keyword
        descriptions.
        """
        return super(piecewise_dcc, self).validate(**kwds)


registered_transforms['dcc'] = piecewise_dcc


class piecewise_cc(TransformedPiecewiseLinearFunction):
    """Discrete CC piecewise representation

    Expresses a piecewise linear function using
    the CC formulation.
    """

    def __init__(self, *args, **kwds):
        super(piecewise_cc, self).__init__(*args, **kwds)

        # create index sets
        polytopes = range(len(self.breakpoints) - 1)
        vertices = range(len(self.breakpoints))

        def vertex_polys(v):
            if v == 0:
                return [v]
            if v == len(self.breakpoints) - 1:
                return [v - 1]
            else:
                return [v - 1, v]

        # create vars
        self.v = variable_dict()
        lmbda = self.v['lambda'] = variable_tuple(variable(lb=0) for v in vertices)
        y = self.v['y'] = variable_tuple(
            variable(domain_type=IntegerSet, lb=0, ub=1) for p in polytopes
        )

        lmbda_tuple = tuple(lmbda)

        # create piecewise constraints
        self.c = constraint_list()

        self.c.append(
            linear_constraint(
                variables=lmbda_tuple + (self.input,),
                coefficients=self.breakpoints + (-1,),
                rhs=0,
            )
        )

        self.c.append(
            linear_constraint(
                variables=lmbda_tuple + (self.output,), coefficients=self.values + (-1,)
            )
        )
        if self.bound == 'ub':
            self.c[-1].lb = 0
        elif self.bound == 'lb':
            self.c[-1].ub = 0
        else:
            assert self.bound == 'eq'
            self.c[-1].rhs = 0

        self.c.append(
            linear_constraint(
                variables=lmbda_tuple, coefficients=(1,) * len(lmbda), rhs=1
            )
        )

        clist = []
        for v in vertices:
            variables = tuple(y[p] for p in vertex_polys(v))
            clist.append(
                linear_constraint(
                    variables=variables + (lmbda[v],),
                    coefficients=(1,) * len(variables) + (-1,),
                    lb=0,
                )
            )
        self.c.append(constraint_tuple(clist))

        self.c.append(
            linear_constraint(variables=tuple(y), coefficients=(1,) * len(y), rhs=1)
        )

    def validate(self, **kwds):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints, values, and
        input variable (e.g., that the list of breakpoints
        is nondecreasing).

        See base class documentation for keyword
        descriptions.
        """
        return super(piecewise_cc, self).validate(**kwds)


registered_transforms['cc'] = piecewise_cc


class piecewise_mc(TransformedPiecewiseLinearFunction):
    """Discrete MC piecewise representation

    Expresses a piecewise linear function using
    the MC formulation.
    """

    def __init__(self, *args, **kwds):
        super(piecewise_mc, self).__init__(*args, **kwds)

        # create indexers
        polytopes = range(len(self.breakpoints) - 1)

        # create constants (using future division)
        # these might also be expressions if the breakpoints
        # or values lists contain mutable objects
        slopes = tuple(
            (self.values[p + 1] - self.values[p])
            / (self.breakpoints[p + 1] - self.breakpoints[p])
            for p in polytopes
        )
        intercepts = tuple(
            self.values[p] - (slopes[p] * self.breakpoints[p]) for p in polytopes
        )

        # create vars
        self.v = variable_dict()
        lmbda = self.v['lambda'] = variable_tuple(variable() for p in polytopes)
        lmbda_tuple = tuple(lmbda)
        y = self.v['y'] = variable_tuple(
            variable(domain_type=IntegerSet, lb=0, ub=1) for p in polytopes
        )
        y_tuple = tuple(y)

        # create piecewise constraints
        self.c = constraint_list()

        self.c.append(
            linear_constraint(
                variables=lmbda_tuple + (self.input,),
                coefficients=(1,) * len(lmbda) + (-1,),
                rhs=0,
            )
        )

        self.c.append(
            linear_constraint(
                variables=lmbda_tuple + y_tuple + (self.output,),
                coefficients=slopes + intercepts + (-1,),
            )
        )
        if self.bound == 'ub':
            self.c[-1].lb = 0
        elif self.bound == 'lb':
            self.c[-1].ub = 0
        else:
            assert self.bound == 'eq'
            self.c[-1].rhs = 0

        clist1 = []
        clist2 = []
        for p in polytopes:
            clist1.append(
                linear_constraint(
                    variables=(y[p], lmbda[p]),
                    coefficients=(self.breakpoints[p], -1),
                    ub=0,
                )
            )
            clist2.append(
                linear_constraint(
                    variables=(lmbda[p], y[p]),
                    coefficients=(1, -self.breakpoints[p + 1]),
                    ub=0,
                )
            )
        self.c.append(constraint_tuple(clist1))
        self.c.append(constraint_tuple(clist2))

        self.c.append(
            linear_constraint(variables=y_tuple, coefficients=(1,) * len(y), rhs=1)
        )

    def validate(self, **kwds):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints, values, and
        input variable (e.g., that the list of breakpoints
        is nondecreasing).

        See base class documentation for keyword
        descriptions.
        """
        ftype = super(piecewise_mc, self).validate(**kwds)
        # this representation does not support step functions
        if ftype == characterize_function.step:
            raise PiecewiseValidationError(
                "The 'mc' piecewise representation does not support step functions."
            )
        return ftype


registered_transforms['mc'] = piecewise_mc


class piecewise_inc(TransformedPiecewiseLinearFunction):
    """Discrete INC piecewise representation

    Expresses a piecewise linear function using
    the INC formulation.
    """

    def __init__(self, *args, **kwds):
        super(piecewise_inc, self).__init__(*args, **kwds)

        # create indexers
        polytopes = range(len(self.breakpoints) - 1)

        # create vars
        self.v = variable_dict()
        delta = self.v['delta'] = variable_tuple(variable() for p in polytopes)
        delta[0].ub = 1
        delta[-1].lb = 0
        delta_tuple = tuple(delta)
        y = self.v['y'] = variable_tuple(
            variable(domain_type=IntegerSet, lb=0, ub=1) for p in polytopes[:-1]
        )

        # create piecewise constraints
        self.c = constraint_list()

        self.c.append(
            linear_constraint(
                variables=(self.input,) + delta_tuple,
                coefficients=(-1,)
                + tuple(
                    self.breakpoints[p + 1] - self.breakpoints[p] for p in polytopes
                ),
                rhs=-self.breakpoints[0],
            )
        )

        self.c.append(
            linear_constraint(
                variables=(self.output,) + delta_tuple,
                coefficients=(-1,)
                + tuple(self.values[p + 1] - self.values[p] for p in polytopes),
            )
        )
        if self.bound == 'ub':
            self.c[-1].lb = -self.values[0]
        elif self.bound == 'lb':
            self.c[-1].ub = -self.values[0]
        else:
            assert self.bound == 'eq'
            self.c[-1].rhs = -self.values[0]

        clist1 = []
        clist2 = []
        for p in polytopes[:-1]:
            clist1.append(
                linear_constraint(
                    variables=(delta[p + 1], y[p]), coefficients=(1, -1), ub=0
                )
            )
            clist2.append(
                linear_constraint(
                    variables=(y[p], delta[p]), coefficients=(1, -1), ub=0
                )
            )
        self.c.append(constraint_tuple(clist1))
        self.c.append(constraint_tuple(clist2))

    def validate(self, **kwds):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints, values, and
        input variable (e.g., that the list of breakpoints
        is nondecreasing).

        See base class documentation for keyword
        descriptions.
        """
        return super(piecewise_inc, self).validate(**kwds)


registered_transforms['inc'] = piecewise_inc


class piecewise_dlog(TransformedPiecewiseLinearFunction):
    """Discrete DLOG piecewise representation

    Expresses a piecewise linear function using the DLOG
    formulation. This formulation uses logarithmic number of
    discrete variables in terms of number of breakpoints.
    """

    def __init__(self, *args, **kwds):
        super(piecewise_dlog, self).__init__(*args, **kwds)

        breakpoints = self.breakpoints
        values = self.values

        if not is_positive_power_of_two(len(breakpoints) - 1):
            raise ValueError(
                "The list of breakpoints must be "
                "of length (2^n)+1 for some positive "
                "integer n. Invalid length: %s" % (len(breakpoints))
            )

        # create branching schemes
        L = log2floor(len(breakpoints) - 1)
        assert 2**L == len(breakpoints) - 1
        B_LEFT, B_RIGHT = self._branching_scheme(L)

        # create indexers
        polytopes = range(len(breakpoints) - 1)
        vertices = range(len(breakpoints))

        def polytope_verts(p):
            return range(p, p + 2)

        # create vars
        self.v = variable_dict()
        lmbda = self.v['lambda'] = variable_dict(
            ((p, v), variable(lb=0)) for p in polytopes for v in polytope_verts(p)
        )
        y = self.v['y'] = variable_tuple(
            variable(domain_type=IntegerSet, lb=0, ub=1) for i in range(L)
        )

        # create piecewise constraints
        self.c = constraint_list()

        self.c.append(
            linear_constraint(
                variables=(self.input,)
                + tuple(lmbda[p, v] for p in polytopes for v in polytope_verts(p)),
                coefficients=(-1,)
                + tuple(breakpoints[v] for p in polytopes for v in polytope_verts(p)),
                rhs=0,
            )
        )

        self.c.append(
            linear_constraint(
                variables=(self.output,)
                + tuple(lmbda[p, v] for p in polytopes for v in polytope_verts(p)),
                coefficients=(-1,)
                + tuple(values[v] for p in polytopes for v in polytope_verts(p)),
            )
        )
        if self.bound == 'ub':
            self.c[-1].lb = 0
        elif self.bound == 'lb':
            self.c[-1].ub = 0
        else:
            assert self.bound == 'eq'
            self.c[-1].rhs = 0

        self.c.append(
            linear_constraint(
                variables=tuple(lmbda.values()), coefficients=(1,) * len(lmbda), rhs=1
            )
        )

        clist = []
        for i in range(L):
            variables = tuple(lmbda[p, v] for p in B_LEFT[i] for v in polytope_verts(p))
            clist.append(
                linear_constraint(
                    variables=variables + (y[i],),
                    coefficients=(1,) * len(variables) + (-1,),
                    ub=0,
                )
            )
        self.c.append(constraint_tuple(clist))
        del clist

        clist = []
        for i in range(L):
            variables = tuple(
                lmbda[p, v] for p in B_RIGHT[i] for v in polytope_verts(p)
            )
            clist.append(
                linear_constraint(
                    variables=variables + (y[i],),
                    coefficients=(1,) * len(variables) + (1,),
                    ub=1,
                )
            )
        self.c.append(constraint_tuple(clist))

    def _branching_scheme(self, L):
        N = 2**L
        B_LEFT = []
        for i in range(1, L + 1):
            start = 1
            step = N // (2**i)
            tmp = []
            while start < N:
                tmp.extend(j - 1 for j in range(start, start + step))
                start += 2 * step
            B_LEFT.append(tmp)

        biglist = range(N)
        B_RIGHT = []
        for i in range(len(B_LEFT)):
            tmp = []
            for j in biglist:
                if j not in B_LEFT[i]:
                    tmp.append(j)
            B_RIGHT.append(sorted(tmp))

        return B_LEFT, B_RIGHT

    def validate(self, **kwds):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints, values, and
        input variable (e.g., that the list of breakpoints
        is nondecreasing).

        See base class documentation for keyword
        descriptions.
        """
        return super(piecewise_dlog, self).validate(**kwds)


registered_transforms['dlog'] = piecewise_dlog


class piecewise_log(TransformedPiecewiseLinearFunction):
    """Discrete LOG piecewise representation

    Expresses a piecewise linear function using the LOG
    formulation. This formulation uses logarithmic number of
    discrete variables in terms of number of breakpoints.
    """

    def __init__(self, *args, **kwds):
        super(piecewise_log, self).__init__(*args, **kwds)

        breakpoints = self.breakpoints
        values = self.values

        if not is_positive_power_of_two(len(breakpoints) - 1):
            raise ValueError(
                "The list of breakpoints must be "
                "of length (2^n)+1 for some positive "
                "integer n. Invalid length: %s" % (len(breakpoints))
            )

        # create branching schemes
        L = log2floor(len(breakpoints) - 1)
        S, B_LEFT, B_RIGHT = self._branching_scheme(L)

        # create indexers
        polytopes = range(len(breakpoints) - 1)
        vertices = range(len(breakpoints))

        # create vars
        self.v = variable_dict()
        lmbda = self.v['lambda'] = variable_tuple(variable(lb=0) for v in vertices)
        y = self.v['y'] = variable_list(
            variable(domain_type=IntegerSet, lb=0, ub=1) for s in S
        )

        # create piecewise constraints
        self.c = constraint_list()

        self.c.append(
            linear_constraint(
                variables=(self.input,) + tuple(lmbda),
                coefficients=(-1,) + breakpoints,
                rhs=0,
            )
        )

        self.c.append(
            linear_constraint(
                variables=(self.output,) + tuple(lmbda), coefficients=(-1,) + values
            )
        )
        if self.bound == 'ub':
            self.c[-1].lb = 0
        elif self.bound == 'lb':
            self.c[-1].ub = 0
        else:
            assert self.bound == 'eq'
            self.c[-1].rhs = 0

        self.c.append(
            linear_constraint(
                variables=tuple(lmbda), coefficients=(1,) * len(lmbda), rhs=1
            )
        )

        clist = []
        for s in S:
            variables = tuple(lmbda[v] for v in B_LEFT[s])
            clist.append(
                linear_constraint(
                    variables=variables + (y[s],),
                    coefficients=(1,) * len(variables) + (-1,),
                    ub=0,
                )
            )
        self.c.append(constraint_tuple(clist))
        del clist

        clist = []
        for s in S:
            variables = tuple(lmbda[v] for v in B_RIGHT[s])
            clist.append(
                linear_constraint(
                    variables=variables + (y[s],),
                    coefficients=(1,) * len(variables) + (1,),
                    ub=1,
                )
            )
        self.c.append(constraint_tuple(clist))

    def _branching_scheme(self, n):
        N = 2**n
        S = range(n)
        G = generate_gray_code(n)
        L = tuple(
            [
                k
                for k in range(N + 1)
                if ((k == 0) or (G[k - 1][s] == 1)) and ((k == N) or (G[k][s] == 1))
            ]
            for s in S
        )
        R = tuple(
            [
                k
                for k in range(N + 1)
                if ((k == 0) or (G[k - 1][s] == 0)) and ((k == N) or (G[k][s] == 0))
            ]
            for s in S
        )
        return S, L, R

    def validate(self, **kwds):
        """
        Validate this piecewise linear function by verifying
        various properties of the breakpoints, values, and
        input variable (e.g., that the list of breakpoints
        is nondecreasing).

        See base class documentation for keyword
        descriptions.
        """
        return super(piecewise_log, self).validate(**kwds)


registered_transforms['log'] = piecewise_log



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


# -*- coding: UTF-8 -*-
"""Tests disjunct fixing."""

import pyomo.common.unittest as unittest
from pyomo.environ import (
    Block,
    Constraint,
    ConcreteModel,
    TransformationFactory,
    NonNegativeReals,
    BooleanVar,
    LogicalConstraint,
    SolverFactory,
    Objective,
    value,
    Var,
    implies,
)
from pyomo.gdp import Disjunct, Disjunction, GDP_Error
from pyomo.opt import check_available_solvers

solvers = check_available_solvers('gurobi')


class TestFixDisjuncts(unittest.TestCase):
    """Tests fixing of disjuncts."""

    def test_fix_disjunct(self):
        """Test for deactivation of trivial constraints."""
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d2 = Disjunct()
        m.d2.c = Constraint()
        m.d = Disjunction(expr=[m.d1, m.d2])
        m.d1.indicator_var.set_value(True)
        m.d2.indicator_var.set_value(False)

        TransformationFactory('gdp.fix_disjuncts').apply_to(m)
        self.assertTrue(m.d1.indicator_var.fixed)
        self.assertTrue(m.d1.active)
        self.assertTrue(m.d2.indicator_var.fixed)
        self.assertFalse(m.d2.active)
        self.assertEqual(m.d1.ctype, Block)
        self.assertEqual(m.d2.ctype, Block)
        self.assertTrue(m.d2.c.active)

    def test_xor_not_sum_to_1(self):
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d2 = Disjunct()
        m.d = Disjunction(expr=[m.d1, m.d2], xor=True)
        m.d1.indicator_var.set_value(True)
        m.d2.indicator_var.set_value(True)
        with self.assertRaises(GDP_Error):
            TransformationFactory('gdp.fix_disjuncts').apply_to(m)

    def test_disjunction_not_sum_to_1(self):
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d2 = Disjunct()
        m.d = Disjunction(expr=[m.d1, m.d2], xor=False)
        m.d1.indicator_var.set_value(False)
        m.d2.indicator_var.set_value(False)
        with self.assertRaises(GDP_Error):
            TransformationFactory('gdp.fix_disjuncts').apply_to(m)

    def test_disjunct_not_binary(self):
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d2 = Disjunct()
        m.d = Disjunction(expr=[m.d1, m.d2])
        m.d1.binary_indicator_var.domain = NonNegativeReals
        m.d2.binary_indicator_var.domain = NonNegativeReals
        m.d1.binary_indicator_var.set_value(0.5)
        m.d2.binary_indicator_var.set_value(0.5)
        with self.assertRaisesRegex(
            GDP_Error,
            "The value of the indicator_var of "
            "Disjunct 'd1' is None. All indicator_vars "
            "must have values before calling "
            "'fix_disjuncts'.",
        ):
            TransformationFactory('gdp.fix_disjuncts').apply_to(m)

    def test_disjuncts_partially_fixed(self):
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d2 = Disjunct()
        m.d = Disjunction(expr=[m.d1, m.d2])
        m.another1 = Disjunct()
        m.another2 = Disjunct()
        m.another = Disjunction(expr=[m.another1, m.another2])

        m.d1.indicator_var.set_value(True)
        m.d2.indicator_var.set_value(False)

        with self.assertRaisesRegex(
            GDP_Error,
            "The value of the indicator_var of "
            "Disjunct 'another1' is None. All indicator_vars "
            "must have values before calling "
            "'fix_disjuncts'.",
        ):
            TransformationFactory('gdp.fix_disjuncts').apply_to(m)

    @unittest.skipIf('gurobi' not in solvers, "Gurobi solver not available")
    def test_logical_constraints_transformed(self):
        """It is expected that the result of this transformation is a MI(N)LP,
        so check that LogicalConstraints are handled correctly"""
        m = ConcreteModel()
        m.x = Var(bounds=(0, 10))
        m.d1 = Disjunct()
        m.d2 = Disjunct()
        m.d2.c = Constraint()
        m.d = Disjunction(expr=[m.d1, m.d2])
        m.another = Disjunction(expr=[[m.x == 3], [m.x == 0]])
        m.Y = BooleanVar()
        m.global_logical = LogicalConstraint(expr=m.Y.xor(m.d1.indicator_var))
        m.d1.logical = LogicalConstraint(
            expr=implies(~m.Y, m.another.disjuncts[0].indicator_var)
        )
        m.obj = Objective(expr=m.x)

        m.d1.indicator_var.set_value(True)
        m.d2.indicator_var.set_value(False)
        m.another.disjuncts[0].indicator_var.set_value(True)
        m.another.disjuncts[1].indicator_var.set_value(False)

        TransformationFactory('gdp.fix_disjuncts').apply_to(m)

        # Make sure there are no active LogicalConstraints
        self.assertEqual(
            len(
                list(
                    m.component_data_objects(
                        LogicalConstraint, active=True, descend_into=(Block, Disjunct)
                    )
                )
            ),
            0,
        )
        # See that it solves as expected
        SolverFactory('gurobi').solve(m)
        self.assertTrue(value(m.d1.indicator_var))
        self.assertFalse(value(m.d2.indicator_var))
        self.assertTrue(value(m.another.disjuncts[0].indicator_var))
        self.assertFalse(value(m.another.disjuncts[1].indicator_var))
        self.assertEqual(value(m.Y.get_associated_binary()), 0)
        self.assertEqual(value(m.x), 3)

    def test_reclassify_deactivated_disjuncts(self):
        m = ConcreteModel()
        m.d = Disjunct([1, 2, 3])
        m.disjunction = Disjunction(expr=[m.d[1], m.d[2], m.d[3]])
        m.d[1].deactivate()
        m.d[2].indicator_var = True
        m.d[3].indicator_var = False

        TransformationFactory('gdp.fix_disjuncts').apply_to(m)

        self.assertTrue(m.d[1].indicator_var.fixed)
        self.assertFalse(value(m.d[1].indicator_var))
        self.assertFalse(m.d[1].active)
        self.assertEqual(m.d[1].ctype, Block)

        self.assertTrue(m.d[2].indicator_var.fixed)
        self.assertTrue(value(m.d[2].indicator_var))
        self.assertTrue(m.d[2].active)
        self.assertTrue(m.d[3].indicator_var.fixed)
        self.assertFalse(value(m.d[3].indicator_var))
        self.assertFalse(m.d[3].active)
        self.assertEqual(m.d[1].ctype, Block)
        self.assertEqual(m.d[2].ctype, Block)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.visitor import StreamBasedExpressionVisitor
from pyomo.core.expr.numeric_expr import (
    ExternalFunctionExpression,
    NegationExpression,
    PowExpression,
    MaxExpression,
    MinExpression,
    ProductExpression,
    MonomialTermExpression,
    DivisionExpression,
    SumExpression,
    Expr_ifExpression,
    UnaryFunctionExpression,
    AbsExpression,
)
from pyomo.core.expr.numvalue import NumericConstant
from pyomo.core.expr.relational_expr import (
    RangedExpression,
    InequalityExpression,
    EqualityExpression,
)
from pyomo.core.base.var import VarData, ScalarVar
from pyomo.core.base.param import ParamData, ScalarParam
from pyomo.core.base.expression import ExpressionData, ScalarExpression
from pyomo.repn.util import ExitNodeDispatcher
from pyomo.common.numeric_types import native_numeric_types
from pyomo.common.collections import ComponentSet


def handle_var(node, collector):
    collector.variables.add(node)
    return None


def handle_param(node, collector):
    collector.params.add(node)
    return None


def handle_named_expression(node, collector):
    collector.named_expressions.add(node)
    return None


def handle_external_function(node, collector):
    collector.external_functions.add(node)
    return None


def handle_skip(node, collector):
    return None


collector_handlers = ExitNodeDispatcher()
collector_handlers[VarData] = handle_var
collector_handlers[ParamData] = handle_param
collector_handlers[ExpressionData] = handle_named_expression
collector_handlers[ScalarExpression] = handle_named_expression
collector_handlers[ExternalFunctionExpression] = handle_external_function
collector_handlers[NegationExpression] = handle_skip
collector_handlers[PowExpression] = handle_skip
collector_handlers[MaxExpression] = handle_skip
collector_handlers[MinExpression] = handle_skip
collector_handlers[ProductExpression] = handle_skip
collector_handlers[MonomialTermExpression] = handle_skip
collector_handlers[DivisionExpression] = handle_skip
collector_handlers[SumExpression] = handle_skip
collector_handlers[Expr_ifExpression] = handle_skip
collector_handlers[UnaryFunctionExpression] = handle_skip
collector_handlers[AbsExpression] = handle_skip
collector_handlers[RangedExpression] = handle_skip
collector_handlers[InequalityExpression] = handle_skip
collector_handlers[EqualityExpression] = handle_skip
collector_handlers[int] = handle_skip
collector_handlers[float] = handle_skip
collector_handlers[NumericConstant] = handle_skip


class _ComponentFromExprCollector(StreamBasedExpressionVisitor):
    def __init__(self, **kwds):
        self.named_expressions = ComponentSet()
        self.variables = ComponentSet()
        self.params = ComponentSet()
        self.external_functions = ComponentSet()
        super().__init__(**kwds)

    def exitNode(self, node, data):
        if type(node) in native_numeric_types:
            # we need this here to handle numpy
            # (we can't put numpy in the dispatcher?)
            return None
        return collector_handlers[node.__class__](node, self)

    def beforeChild(self, node, child, child_idx):
        if child in self.named_expressions:
            return False, None
        return True, None


_visitor = _ComponentFromExprCollector()


def collect_components_from_expr(expr):
    _visitor.__init__()
    _visitor.walk_expression(expr)
    return (
        _visitor.named_expressions,
        _visitor.variables,
        _visitor.params,
        _visitor.external_functions,
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.common.deprecation import deprecated
from pyomo.core.base import (
    Transformation,
    TransformationFactory,
    Var,
    Constraint,
    Objective,
    minimize,
    NonNegativeReals,
    NonPositiveReals,
    Reals,
    Block,
    Model,
    ConcreteModel,
)
from pyomo.duality.collect import collect_linear_terms


def load():
    pass


logger = logging.getLogger('pyomo.core')


#
# This transformation creates a new block that
# is the dual of the specified block.  If no block is
# specified, then the entire model is dualized.
# This returns a new Block object.
#
@TransformationFactory.register(
    'duality.linear_dual', doc="[DEPRECATED] Dualize a linear model"
)
@deprecated(
    "Use of the pyomo.duality package is deprecated. There are known bugs "
    "in pyomo.duality, and we do not recommend the use of this code. "
    "Development of dualization capabilities has been shifted to "
    "the Pyomo Adversarial Optimization (PAO) library. Please contact "
    "William Hart for further details (wehart@sandia.gov).",
    version='5.6.2',
)
class LinearDual_PyomoTransformation(Transformation):
    def __init__(self):
        super(LinearDual_PyomoTransformation, self).__init__()

    def _create_using(self, instance, **kwds):
        options = kwds.pop('options', {})
        bname = options.get('block', None)
        #
        # Iterate over the model collecting variable data,
        # until the block is found.
        #
        block = None
        if block is None:
            block = instance
        else:
            for name, data in instance.component_map(Block, active=True).items():
                if name == bname:
                    block = instance
        if block is None:
            raise RuntimeError("Missing block: " + bname)
        #
        # Generate the dual
        #
        instance_ = self._dualize(block)

        return instance_

    def _dualize(self, block, unfixed=[]):
        """
        Generate the dual of a block
        """
        #
        # Collect linear terms from the block
        #
        A, b_coef, c_rhs, c_sense, d_sense, vnames, cnames, v_domain = (
            collect_linear_terms(block, unfixed)
        )
        ##print(A)
        ##print(vnames)
        ##print(cnames)
        ##print(list(A.keys()))
        ##print("---")
        ##print(A.keys())
        ##print(c_sense)
        ##print(c_rhs)
        #
        # Construct the block
        #
        if isinstance(block, Model):
            dual = ConcreteModel()
        else:
            dual = Block()
        dual.construct()
        _vars = {}

        def getvar(name, ndx=None):
            v = _vars.get((name, ndx), None)
            if v is None:
                v = Var()
                if ndx is None:
                    v_name = name
                elif type(ndx) is tuple:
                    v_name = "%s[%s]" % (name, ','.join(map(str, ndx)))
                else:
                    v_name = "%s[%s]" % (name, str(ndx))
                setattr(dual, v_name, v)
                _vars[name, ndx] = v
            return v

        #
        # Construct the objective
        #
        if d_sense == minimize:
            dual.o = Objective(
                expr=sum(
                    -b_coef[name, ndx] * getvar(name, ndx) for name, ndx in b_coef
                ),
                sense=d_sense,
            )
        else:
            dual.o = Objective(
                expr=sum(b_coef[name, ndx] * getvar(name, ndx) for name, ndx in b_coef),
                sense=d_sense,
            )
        #
        # Construct the constraints
        #
        for cname in A:
            for ndx, terms in A[cname].items():
                expr = 0
                for term in terms:
                    expr += term.coef * getvar(term.var, term.ndx)
                if not (cname, ndx) in c_rhs:
                    c_rhs[cname, ndx] = 0.0
                if c_sense[cname, ndx] == 'e':
                    e = expr - c_rhs[cname, ndx] == 0
                elif c_sense[cname, ndx] == 'l':
                    e = expr - c_rhs[cname, ndx] <= 0
                else:
                    e = expr - c_rhs[cname, ndx] >= 0
                c = Constraint(expr=e)
                if ndx is None:
                    c_name = cname
                elif type(ndx) is tuple:
                    c_name = "%s[%s]" % (cname, ','.join(map(str, ndx)))
                else:
                    c_name = "%s[%s]" % (cname, str(ndx))
                setattr(dual, c_name, c)
            #
            for (name, ndx), domain in v_domain.items():
                v = getvar(name, ndx)
                flag = type(ndx) is tuple and (ndx[-1] == 'lb' or ndx[-1] == 'ub')
                if domain == 1:
                    v.domain = NonNegativeReals
                elif domain == -1:
                    v.domain = NonPositiveReals
                else:
                    # TODO: verify that this case is possible
                    v.domain = Reals

        return dual



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.cp.interval_var import (
    IntervalVar,
    IntervalVarTimePoint,
    IntervalVarLength,
    IntervalVarPresence,
)
from pyomo.core.expr import GetItemExpression, GetAttrExpression
from pyomo.environ import ConcreteModel, Integers, Reference, Set, value, Var


class TestScalarIntervalVar(unittest.TestCase):
    def test_initialize_with_no_data(self):
        m = ConcreteModel()
        m.i = IntervalVar()

        self.assertIsInstance(m.i.start_time, IntervalVarTimePoint)
        self.assertEqual(m.i.start_time.domain, Integers)
        self.assertIsNone(m.i.start_time.lower)
        self.assertIsNone(m.i.start_time.upper)

        self.assertIsInstance(m.i.end_time, IntervalVarTimePoint)
        self.assertEqual(m.i.end_time.domain, Integers)
        self.assertIsNone(m.i.end_time.lower)
        self.assertIsNone(m.i.end_time.upper)

        self.assertIsInstance(m.i.length, IntervalVarLength)
        self.assertEqual(m.i.length.domain, Integers)
        self.assertIsNone(m.i.length.lower)
        self.assertIsNone(m.i.length.upper)

        self.assertIsInstance(m.i.is_present, IntervalVarPresence)

    def test_add_components_that_do_not_belong(self):
        m = ConcreteModel()
        m.i = IntervalVar()

        with self.assertRaisesRegex(
            ValueError,
            "Attempting to declare a block component using the name of a "
            "reserved attribute:\n\tnew_thing",
        ):
            m.i.new_thing = IntervalVar()

    def test_start_and_end_bounds(self):
        m = ConcreteModel()
        m.i = IntervalVar(start=(0, 5))
        self.assertEqual(m.i.start_time.lower, 0)
        self.assertEqual(m.i.start_time.upper, 5)

        m.i.end_time.bounds = (12, 14)

        self.assertEqual(m.i.end_time.lower, 12)
        self.assertEqual(m.i.end_time.upper, 14)

    def test_constant_length_and_start(self):
        m = ConcreteModel()
        m.i = IntervalVar(length=7, start=3)

        self.assertEqual(m.i.length.lower, 7)
        self.assertEqual(m.i.length.upper, 7)

        self.assertEqual(m.i.start_time.lower, 3)
        self.assertEqual(m.i.start_time.upper, 3)

    def test_non_optional(self):
        m = ConcreteModel()
        m.i = IntervalVar(length=2, end=(4, 9), optional=False)

        self.assertEqual(value(m.i.is_present), True)
        self.assertTrue(m.i.is_present.fixed)
        self.assertFalse(m.i.optional)

        # Should also be true by default
        m.i2 = IntervalVar()

        self.assertEqual(value(m.i2.is_present), True)
        self.assertTrue(m.i.is_present.fixed)
        self.assertFalse(m.i2.optional)

    def test_optional(self):
        m = ConcreteModel()
        m.i = IntervalVar(optional=True)

        self.assertFalse(m.i.is_present.fixed)
        self.assertTrue(m.i.optional)

        # Now set to False
        m.i.optional = False
        self.assertEqual(value(m.i.is_present), True)
        self.assertTrue(m.i.is_present.fixed)
        self.assertFalse(m.i.optional)

    def test_is_present_fixed_False(self):
        m = ConcreteModel()
        m.i = IntervalVar(optional=True)

        m.i.is_present.fix(False)
        self.assertTrue(m.i.optional)


class TestIndexedIntervalVar(unittest.TestCase):
    def test_initialize_with_no_data(self):
        m = ConcreteModel()

        m.i = IntervalVar([1, 2])

        for j in [1, 2]:
            self.assertIsInstance(m.i[j].start_time, IntervalVarTimePoint)
            self.assertEqual(m.i[j].start_time.domain, Integers)
            self.assertIsNone(m.i[j].start_time.lower)
            self.assertIsNone(m.i[j].start_time.upper)

            self.assertIsInstance(m.i[j].end_time, IntervalVarTimePoint)
            self.assertEqual(m.i[j].end_time.domain, Integers)
            self.assertIsNone(m.i[j].end_time.lower)
            self.assertIsNone(m.i[j].end_time.upper)

            self.assertIsInstance(m.i[j].length, IntervalVarLength)
            self.assertEqual(m.i[j].length.domain, Integers)
            self.assertIsNone(m.i[j].length.lower)
            self.assertIsNone(m.i[j].length.upper)

            self.assertIsInstance(m.i[j].is_present, IntervalVarPresence)

    def test_constant_length(self):
        m = ConcreteModel()
        m.i = IntervalVar(['a', 'b'], length=45)

        for j in ['a', 'b']:
            self.assertEqual(m.i[j].length.lower, 45)
            self.assertEqual(m.i[j].length.upper, 45)

    def test_rule_based_start(self):
        m = ConcreteModel()

        def start_rule(m, i):
            return (1 - i, 13 + i)

        m.act = IntervalVar([1, 2, 3], start=start_rule, length=4)

        for i in [1, 2, 3]:
            self.assertEqual(m.act[i].start_time.lower, 1 - i)
            self.assertEqual(m.act[i].start_time.upper, 13 + i)

            self.assertEqual(m.act[i].length.lower, 4)
            self.assertEqual(m.act[i].length.upper, 4)

            self.assertFalse(m.act[i].optional)
            self.assertTrue(m.act[i].is_present.fixed)
            self.assertEqual(value(m.act[i].is_present), True)

    def test_optional(self):
        m = ConcreteModel()
        m.act = IntervalVar([1, 2], end=[0, 10], optional=True)

        for i in [1, 2]:
            self.assertTrue(m.act[i].optional)
            self.assertFalse(m.act[i].is_present.fixed)

            self.assertEqual(m.act[i].end_time.lower, 0)
            self.assertEqual(m.act[i].end_time.upper, 10)

        # None doesn't make sense for this:
        with self.assertRaisesRegex(
            ValueError, "Cannot set 'optional' to None: Must be True or False."
        ):
            m.act[1].optional = None

        # We can change it, and that has the correct effect on is_present
        m.act[1].optional = False
        self.assertFalse(m.act[1].optional)
        self.assertTrue(m.act[1].is_present.fixed)

        m.act[1].optional = True
        self.assertTrue(m.act[1].optional)
        self.assertFalse(m.act[1].is_present.fixed)

    def test_optional_rule(self):
        m = ConcreteModel()
        m.idx = Set(initialize=[(4, 2), (5, 2)], dimen=2)

        def optional_rule(m, i, j):
            return i % j == 0

        m.act = IntervalVar(m.idx, optional=optional_rule)
        self.assertTrue(m.act[4, 2].optional)
        self.assertFalse(m.act[5, 2].optional)

    def test_index_by_expr(self):
        m = ConcreteModel()
        m.act = IntervalVar([(1, 2), (2, 1), (2, 2)])
        m.i = Var(domain=Integers)
        m.i2 = Var([1, 2], domain=Integers)

        thing1 = m.act[m.i, 2]
        self.assertIsInstance(thing1, GetItemExpression)
        self.assertEqual(len(thing1.args), 3)
        self.assertIs(thing1.args[0], m.act)
        self.assertIs(thing1.args[1], m.i)
        self.assertEqual(thing1.args[2], 2)

        thing2 = thing1.start_time
        self.assertIsInstance(thing2, GetAttrExpression)
        self.assertEqual(len(thing2.args), 2)
        self.assertIs(thing2.args[0], thing1)
        self.assertEqual(thing2.args[1], 'start_time')

        expr1 = m.act[m.i, 2].start_time.before(m.act[m.i**2, 1].end_time)

    def test_reference(self):
        m = ConcreteModel()
        m.act = IntervalVar([1, 2], end=[0, 10], optional=True)

        thing = Reference(m.act[:].is_present)
        self.assertIs(thing[1], m.act[1].is_present)
        self.assertIs(thing[2], m.act[2].is_present)

        thing = Reference(m.act[:].start_time)
        self.assertIs(thing[1], m.act[1].start_time)
        self.assertIs(thing[2], m.act[2].start_time)

        thing = Reference(m.act[:].end_time)
        self.assertIs(thing[1], m.act[1].end_time)
        self.assertIs(thing[2], m.act[2].end_time)

        thing = Reference(m.act[:].length)
        self.assertIs(thing[1], m.act[1].length)
        self.assertIs(thing[2], m.act[2].length)
